import copy
import logging
import numpy as np
import os.path as osp
import schedula as sh

log = logging.getLogger(__name__)
dsp = sh.BlueDispatcher(name='load', description='Read/merge/parse input data.')


def check_ext(fpath, *args, ext=('xls', 'xlsx')):
    return osp.splitext(fpath)[1][1:] in ext


@sh.add_function(dsp, outputs=['raw_data'], input_domain=check_ext)
def read_excel(input_path):
    """
    Read input file.

    :param input_path:
        Input file path.
    :type input_path: str

    :return:
        Raw data of input file.
    :rtype: dict
    """
    import xlrd
    import pandas as pd
    raw_data = {}
    with pd.ExcelFile(input_path) as xl:
        for k in ('inputs', 'config', 'vehicle_inputs', 'time_series'):
            try:
                if k == 'time_series':
                    raw_data[k] = xl.parse(k).dropna(1).to_dict(orient='list')
                else:
                    df = xl.parse(k, header=None, index_col=0, usecols=[0, 1]).T
                    raw_data[k] = df.dropna(1).to_dict(orient='records')[0]
            except IndexError:
                log.warning('Sheet `%s` is not well formatted!' % k)
            except xlrd.biffh.XLRDError:
                log.warning('Missing sheet (`%s`)!' % k)

    return raw_data


dsp.add_data('raw_data', {}, sh.inf(1, 0))


@sh.add_function(dsp, outputs=['vehicle_id'])
def get_vehicle_id(raw_data):
    """
    Get vehicle ID from raw data.

    :param raw_data:
        Raw data of input file.
    :type raw_data: dict

    :return:
        Vehicle ID.
    :rtype: int
    """
    return raw_data['config']['vehicle_id']


dsp.add_data(
    'db_path', osp.join(osp.dirname(__file__), 'db', 'EuroSegmentCar_cleaned.csv'),
    sh.inf(1, 0)
)


@sh.add_function(dsp, outputs=['db_path'])
def get_db_path(raw_data):
    """
    Get data base file path from raw data.

    :param raw_data:
        Raw data of input file.
    :type raw_data: dict

    :return:
        Data base file path.
    :rtype: str
    """
    return raw_data.get('config', {}).get('db_path', sh.NONE)


_db_map = {
    "Transmission  / Gear ratio-Final drive": 'final_drive_ratio',
    "Transmission  / Gear ratio-Gear Box Ratios": "gear_box_ratios",
    'Weights-Empty mass': 'vehicle_mass',
    'Performance-Top speed': 'vehicle_max_speed',
    'General Specifications-Carbody': 'type_of_car',
    'Exterior sizes-Width': 'car_width',
    'Exterior sizes-Height': 'car_height',
    'Weights-Unladen mass': 'kerb_weight',
    'Exterior sizes-Wheelbase': 'wheelbase',
    'Drive-Wheel drive': 'car_type', 'Drive-Fuel': 'fuel_type',
    'Chassis-Rolling Radius Dynamic': 'r_dynamic',
    'Fuel Engine-Max torque': 'engine_max_torque',
    'Fuel Engine-Stroke': 'fuel_engine_stroke',
    'Drive-Total max power': 'max_power',
    'Fuel Engine-Turbo': 'fuel_turbo',
    'Fuel Engine-Capacity': 'fuel_eng_capacity',
    'General Specifications-Transmission': 'gearbox_type',
    'Fuel Engine-Max power': 'engine_max_power',
    'Electric Engine-Total max power': 'motor_max_power',
    'Electric Engine-Max torque': 'motor_max_torque',
    'Chassis-Rolling Radius Static': 'tyre_radius',
    "Fuel Engine-Max power RPM": "engine_max_speed_at_max_power"
}


@sh.add_function(dsp, outputs=['vehicle_db'])
def load_vehicle_db(db_path):
    """
    Load vehicle data base.

    :param db_path:
        Data base file path.
    :type db_path: str

    :return:
        Vehicle database.
    :rtype: dict
    """
    import pandas as pd
    df = pd.read_csv(db_path, encoding="ISO-8859-1", index_col=0)
    df = df[list(_db_map)].rename(columns=_db_map)

    df['gear_box_ratios'] = df['gear_box_ratios'].fillna('[]').apply(
        lambda x: [float(v) for v in x[1:-1].split('-') if v != '']
    )
    df.loc[df['fuel_type'] == 'petrol', 'ignition_type'] = 'positive'
    df.loc[df['fuel_type'] == 'diesel', 'ignition_type'] = 'compression'
    b = df['fuel_type'] == 'electricity'
    c = df['fuel_type'] != 'electricity'
    df.loc[b, ['ignition_type']] = 'electricity'
    df.loc[b, ['gear_box_ratios']] = 1
    df.loc[b, ['tyre_radius']] /= 1000  # meters.
    df.loc[c, ['tyre_radius']] = (24.5 / 2 * 2.54) / 100  # meters.
    df['driveline_slippage'] = 0

    b = df['gearbox_type'] == 'automatic'
    # b |= df['gearbox_type'] == 'single-speed fixed gear'
    df['transmission'] = np.where(b, 'automatic', 'manual')
    df['driveline_efficiency'] = np.where(b, .9, .93)

    df.loc[(df['fuel_type'] == 'electricity') & (df['gearbox_type'] == 'single-speed fixed gear'), 'driveline_efficiency'] = 0.93
    df.loc[(df['fuel_type'] == 'electricity') & (df['gearbox_type'] != 'single-speed fixed gear'), 'driveline_efficiency'] = 0.93

    df['vehicle_max_speed'] = (df['vehicle_max_speed'] / 3.6).values.astype(int)
    df.round({'vehicle_max_speed': 2})
    df['type_of_car'] = df["type_of_car"].str.strip()
    r = np.where(df['car_type'] == 'front', 2, 6)
    r[df['car_type'] == 'rear'] = 4
    df['car_type'] = r

    b = df['ignition_type'] == 'positive'
    df['idle_engine_speed_median'] = np.where(b, 750, 850)
    df['idle_engine_speed_std'] = 50
    df['r_dynamic'] /= 1000

    return df.to_dict('index')


@sh.add_function(dsp, outputs=['vehicle_inputs'])
def get_vehicle_inputs(vehicle_id, vehicle_db):
    """
    Get vehicle data.

    :param vehicle_id:
        Vehicle ID.
    :type vehicle_id: int

    :param vehicle_db:
        Vehicle database.
    :type vehicle_db: dict

    :return:
        Vehicle inputs.
    :rtype: dict
    """
    return vehicle_db[vehicle_id]


dsp.add_data('inputs', {}, sh.inf(2, 0))


@sh.add_function(dsp, outputs=['data'])
def merge_data(vehicle_inputs, raw_data, inputs):
    """
    Merge data.

    :param vehicle_inputs:
        Vehicle inputs.
    :type vehicle_inputs: dict

    :param raw_data:
        Raw data of input file.
    :type raw_data: dict

    :param inputs:
        User inputs.
    :type inputs:

    :return:
        Merged data.
    :rtype: dict
    """
    d = {'vehicle_inputs': vehicle_inputs}
    d = sh.combine_nested_dicts(d, raw_data, inputs, depth=2)
    return sh.combine_dicts(
        d.pop('time_series', {}), d.pop('vehicle_inputs', {}),
        d.get('inputs', {})
    )


def format_data(data):
    """
    Format data.

    :param data:
        Data to be formatted.
    :type data: dict

    :return:
        Formatted data.
    :rtype: dict
    """
    data = copy.deepcopy(data)
    if isinstance(data.get('gear_box_ratios'), str):
        import json
        data['gear_box_ratios'] = json.loads(data['gear_box_ratios'])
    for k, v in list(data.items()):
        if isinstance(v, str):
            if v:
                continue
        elif not np.atleast_1d(np.isnan(v)).any():
            continue
        data.pop(k)
    return data


dsp.add_data('data', filters=[format_data])

if __name__ == '__main__':
    dsp.register(memo={}).plot()



from os import path as osp, chdir
from tabulate import tabulate
from co2mpas_driver import load as ld

my_dir = osp.dirname(osp.abspath(__file__))
chdir(my_dir)


def print_vehicle_db():
    """
    Sample file printing an indicative set of columns from the dataset for all vehicles of a Euro-Segment.

    :return:
    """
    # Vehicle databased based on the Euro Car Segment classification
    db_path = osp.abspath(osp.join(my_dir, 'db', 'EuroSegmentCar.csv'))
    import pandas as pd
    df = pd.read_csv(db_path, encoding="ISO-8859-1", index_col=0)
    df = df[list(ld._db_map)].rename(columns=ld._db_map)
    df_ = df.loc[:, ld._db_map.values()]
    print(tabulate(df_, headers='keys', tablefmt='psql'))


if __name__ == '__main__':
    print_vehicle_db()


from co2mpas_driver import dsp as driver
import matplotlib.pyplot as plt
import numpy as np


def simple_run():
    """
    Test vehicle simulation.

    :return:
    """
    veh_ids = [39393, 8188, 40516, 35452, 40225, 7897, 7972, 41388, 5766,
               9645, 9639, 5798, 8280, 34271, 34265, 6378, 39723, 34092, 2592,
               5635, 5630, 7661, 7683]
    v_des = 124/3.6
    v_start = 0
    dt = 0.1
    times = np.arange(0, 100, dt)

    vehicles = [driver(dict(vehicle_id=i, inputs=dict(inputs=dict(
        gear_shifting_style=0.8, driver_style=0.8, starting_velocity=0,
        duration=100, sim_start=0, sim_step=dt, use_linear_gs=True,
        use_cubic=False))))['outputs']['driver_simulation_model'] for i in
                veh_ids]
    res = {}
    for myt in times:
        for my_veh in vehicles:
            if myt == times[0]:
                my_veh.reset(v_start)
                res[my_veh] = {'accel': [0], 'speed': [v_start], 'gear': [0]}
                continue

            gear, gear_count, next_velocity, acc = my_veh(dt, v_des)
            res[my_veh]['accel'].append(acc)
            res[my_veh]['speed'].append(next_velocity)
            res[my_veh]['gear'].append(gear)

    for my_veh in vehicles:
        plt.figure('Acceleration-Speed')
        plt.plot(res[my_veh]['speed'], res[my_veh]['accel'])
        plt.show()
    return 0


if __name__ == '__main__':
    simple_run()


# -*- coding: utf-8 -*-
#
# Copyright 2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions to simulate free-flow movement of vehicle's.
"""


def gear_for_speed_profiles(gs, curr_speed, current_gear, gear_cnt, sim_step=0.1, clutch_duration=0.5):
    """
    Return the gear that must be used and the clutch condition.

    :param gs:
        Gear limits.
    :type gs: list

    :param curr_speed:
        Current speed.
    :type curr_speed: int

    :param current_gear:
        Current speed.
    :type current_gear: int

    :param gear_cnt:
        Gear count.
    :type gear_cnt: int

    :param sim_step:
        Simulation step. [s]
    :type sim_step: float

    :param clutch_duration:
        Clutch duration in sim step
    :type clutch_duration: int

    :return:
        Current gear & gear count
    :rtype: int, int
    """

    # Model buffer for up shifting and down shifting.
    upshift_offs = 0.0
    downshift_off = 0.1

    gear_limits = [0]
    gear_limits.extend(gs)
    gear_limits.append(200)

    if gear_limits[current_gear - 1] - gear_limits[current_gear - 1] * downshift_off <= curr_speed < \
            gear_limits[current_gear] + gear_limits[current_gear] * upshift_offs:
        if gear_cnt == 0:
            return current_gear, gear_cnt
        else:
            gear_cnt -= 1
            return current_gear, gear_cnt
    else:
        itr = 1
        gear_search = 1
        while itr == 1:
            if gear_limits[gear_search - 1] <= curr_speed < gear_limits[gear_search]:
                gear_cnt = int(divmod(10 * clutch_duration, 10 * sim_step)[0])  # in simulation steps for 0.5 second
                current_gear = gear_search
                itr = 0
            else:
                gear_search += 1
        return current_gear, gear_cnt


def accMFC(velocity, driver_style, desired_velocity, acc_p_curve, dec_p_curve):
    """
    Calculate the MFC free flow acceleration.

    :param velocity:
        speed. (m/s)
    :type velocity: int

    :param driver_style:
        Driver style from 0-1.
    :type driver_style: int

    :param desired_velocity:
        desired velocity (m/s)
    :type desired_velocity: int

    :param acc_p_curve:
        Acceleration potential curve of the gear in use.
    :type acc_p_curve:

    :param dec_p_curve:
        Deceleration potential curve of the gear in use.
    :type dec_p_curve:

    :return:
    """
    if velocity >= desired_velocity:
        on_off = (1 - pow(1 - (velocity - desired_velocity) / 50, 100))
        acc = dec_p_curve(velocity) * driver_style * on_off
    elif velocity >= 0.5 * desired_velocity:
        on_off = (1 - pow(1 + (velocity - desired_velocity) / 50, 100))
        acc = acc_p_curve(velocity) * driver_style * on_off
    else:
        on_off = (1 - 0.8 * pow(1 - velocity / desired_velocity, 60))
        acc = acc_p_curve(velocity) * driver_style * on_off

    return acc


def correct_acc_clutch_on(gear_count, acc, transmission):
    """
    Get the acceleration If clutch is on. Maximum acceleration is
    decreased depending on the transmission.

    :param gear_count:
        Gear count.
    :type gear_count: int

    :param acc:
        Acceleration. (m/s2)
    :type acc:

    :param transmission:
        Transmission type.
    :type transmission: str

    :return:
        Acceleration when clutch on. (m/s2)
    :rtype:
    """

    if gear_count > 0:

        if transmission == 'manual':
            return 0.
        else:
            return acc * 2 / 3
    else:
        return acc



from os import path as osp, chdir
import numpy as np
import matplotlib.pyplot as plt
from co2mpas_driver import dsp as driver
from scipy.interpolate import interp1d

my_dir = osp.dirname(osp.abspath(__file__))
chdir(my_dir)


def simple_run():
    """
     This example computes and plots the driving trajectories with the varying desired speed.

    """
    # A sample car id from the database = {
    car_id_db = {
        'fuel_engine': [35135, 39393, 27748, 8188, 40516, 35452, 40225, 7897, 7972, 41388, 5766, 9645, 9639, 5798, 8280, 34271, 34265, 6378, 39723, 34092, 2592, 5635, 5630, 7661, 7683, 8709, 9769, 1872, 10328, 35476, 41989, 26799, 26851, 27189, 23801, 3079, 36525, 47766, 6386, 33958, 33986, 5725, 5718, 36591, 4350, 39396, 40595, 5909, 5897, 5928, 5915, 40130, 42363, 34760, 34766, 1835, 36101, 42886, 1431, 46547, 44799, 41045, 39820, 34183, 34186, 20612, 20605, 1324, 9882, 4957, 5595, 18831, 18833, 9575, 5380, 9936, 7995, 6331, 18173, 34286, 34279, 20706, 34058, 34057, 24268, 19028, 19058, 7979, 22591, 34202, 40170, 44599, 5358, 5338, 34015, 9872, 9856, 6446, 8866, 9001, 9551, 6222], 
        'electric_engine': [47844]
    }
    car_id = 47844

    # The desired speed
    vdes = 124 / 3.6

    # Current speed
    v_start = 0

    # The simulation step in seconds
    sim_step = 0.1

    # The driving style as described in the TRR paper.
    driver_style = 0.6

    # Duration of the simulation in seconds.
    duration = 300

    # sample time series
    times = np.arange(0, duration + sim_step, sim_step)

    # core model, this will select and execute the proper functions for the given inputs and returns the output
    # You can also pass vehicle database path db_path='path to vehicle db'
    if car_id in car_id_db['electric_engine']:
        my_veh = driver(dict(vehicle_id=car_id, inputs=dict(inputs=dict(
            starting_velocity=v_start, driver_style=driver_style,
            sim_start=0, sim_step=sim_step, duration=duration, use_linear_gs=True,
            use_cubic=False))))['outputs']['driver_simulation_model']
    else:
        # The gear shifting style as described in the TRR paper.
        gs_style = 0.8
        my_veh = driver(dict(vehicle_id=car_id, inputs=dict(inputs=dict(
            gear_shifting_style=gs_style, starting_velocity=v_start, 
            driver_style=driver_style, sim_start=0, sim_step=sim_step, 
            duration=duration, degree=4, use_linear_gs=True, 
            use_cubic=False))))['outputs']['driver_simulation_model']
    # Plots workflow of the core model, this will automatically open an internet browser and show the work flow
    # of the core model. you can click all the rectangular boxes to see in detail sub models like load, model,
    # write and plot.
    # you can uncomment to plot the work flow of the model
    # driver.plot(1)  # Note: run your IDE as Administrator if file permission error.

    f_des = interp1d(
        [0,     210,    700,    1050,   1400,   3150,   4550,   5250],  # Position (m)
        [20,     20,     30,     5,      25,     35,     15,     0],    # Desired speed (m/s)
        kind='next')
    res = {}
    for myt in times:
        if myt == times[0]:
            my_veh.reset(v_start)
            res = {'accel': [0], 'speed': [v_start],
                   'position': [0], 'gear': [0], 'v_des': [f_des(0)], 'fc': [0]}
            continue
        res['v_des'].append(f_des(res['position'][-1]))
        gear, gear_count, next_velocity, acc = my_veh(sim_step, res['v_des'][-1])

        if car_id in car_id_db['electric_engine']:
            fc = 0
        else:
            fc = my_veh.calculate_fuel_consumption(next_velocity, acc, gear, gear_count, sim_step)
            
        res['accel'].append(acc)
        res['speed'].append(next_velocity)
        res['gear'].append(gear)
        res['position'].append(res['position'][-1] + next_velocity * sim_step)
        res['fc'].append(fc)

    fig = plt.figure('Time-Speed')
    plt.plot(times, res['speed'], label=f'Simulation car:{car_id}, DS: {driver_style}')
    plt.plot(times, res['v_des'], label=f'Desired speed (m/s)')
    plt.legend()
    fig.suptitle('Speed over time', x=0.54, y=1, fontsize=12)
    plt.xlabel('Time (s)', fontsize=14)
    plt.ylabel('Speed (m/s)', fontsize=12)
    plt.grid()

    fig = plt.figure('Time-Acceleration')
    plt.plot(times, res['accel'], label=f'Simulation car:{car_id}, DS: {driver_style}')
    plt.legend()
    fig.suptitle('Acceleration over time', x=0.54, y=1, fontsize=12)
    plt.xlabel('Time (s)', fontsize=14)
    plt.ylabel('Acceleration (m/s2)', fontsize=12)
    plt.grid()

    if car_id in car_id_db['electric_engine']:
        pass
    else:
        fig = plt.figure('Time-Fuel Consumption')
        plt.plot(times, np.dot(res['fc'], sim_step), label=f'Simulation car:{car_id}, DS: {driver_style}')
        plt.legend()
        fig.suptitle('Fuel Consumption over time', x=0.54, y=1, fontsize=12)
        plt.xlabel('Time (s)', fontsize=14)
        plt.ylabel('Fuel consumption rate (g/s)', fontsize=12)
        plt.grid()

    fig = plt.figure('Distance-Speed')
    plt.plot(res['position'], res['speed'], label=f'Simulation car:{car_id}, DS: {driver_style}')
    plt.plot(res['position'], res['v_des'], label=f'Desired speed (m/s)')
    plt.legend()
    fig.suptitle('Speed over distance', x=0.54, y=1, fontsize=12)
    plt.xlabel('Distance (m)', fontsize=14)
    plt.ylabel('Speed (m/s)', fontsize=12)
    plt.grid()

    fig = plt.figure('Distance-Acceleration')
    plt.plot(res['position'], res['accel'], label=f'Simulation car:{car_id}, DS: {driver_style}')
    plt.legend()
    fig.suptitle('Acceleration over distance', x=0.54, y=1, fontsize=12)
    plt.xlabel('Distance (m)', fontsize=14)
    plt.ylabel('Acceleration (m/s2)', fontsize=12)
    plt.grid()
    plt.show()


if __name__ == '__main__':
    simple_run()



import matplotlib.pyplot as plt
import numpy as np


def plot_speed_acceleration_from_coefs(poly_coefs_per_gear, speeds_per_gear, accelerations_per_gear):
    """
    Plot the speed acceleration diagram created.

    :param poly_coefs_per_gear: Polynomial coefficients for the acceleration over speed curves
    :param speeds_per_gear: Speed points
    :param accelerations_per_gear: Acceleration points
    :return:
    """

    degree = len(poly_coefs_per_gear[0]) - 1
    vars = np.arange(degree, -1, -1)

    plt.figure('speed acceleration regression results of degree = ' + str(degree))

    for speeds, acceleration, fit_coef in zip(speeds_per_gear, accelerations_per_gear, poly_coefs_per_gear):
        plt.plot(speeds, acceleration, 'kx')

        # x_new = np.linspace(speeds[0], speeds[-1], 100)
        x_new = np.arange(speeds[0], speeds[-1], 0.1)
        a_new = np.array([np.dot(fit_coef, np.power(i, vars)) for i in x_new])

        plt.plot(x_new, a_new, 'rx')


import numpy as np


def find_car_gear(my_car, speed, rpm):
    """
    Use speed and rpm to estimate the current gear.
    :param my_car: Car specs. type:Dict
    :param speed: Car speed m/s
    :param rpm: Engine RPM
    :return: Integer corresponding to the current gear.
    """

    gr_str = np.array([1 / float(i) / my_car.final_drive for i in my_car.gr])

    ratio = speed * 60 / my_car.wheelbase / rpm

    res = np.abs(gr_str - ratio)

    return res.argmin()


def find_list_of_tans_from_coefs(coefs_per_gear, Start, Stop):
    """

    Gets coefficients and speed boundaries and returns Tans value for per speed
    per gear

    :param coefs_per_gear:
    :param Start:
    :param Stop:
    :return:
    """
    degree = len(coefs_per_gear[0]) - 1
    vars = np.arange(degree, -1, -1)

    Tans = []

    for gear, coefs in enumerate(coefs_per_gear):
        x_new = np.arange(Start[gear], Stop[gear], 0.1)
        a_new = np.array([np.dot(coefs, np.power(i, vars)) for i in x_new])

        Tans.append(np.diff(a_new) * 10)

    return Tans


def find_gs_cut_tans(tmp_min, tmp_max, tan, tmp_min_next, gs_style):
    """

    Find where gear is changed, based on tans and gs_style

    :param tmp_min:
    :param tmp_max:
    :param tan:
    :param tmp_min_next:
    :param cutoff:
    :return:
    """
    max_tan = np.max(tan)
    min_tan = np.min(tan)
    acc_range = max_tan - min_tan

    # tan starts from positive and goes negative, so I use (1 - cutoff)
    # for the percentage
    if gs_style > 0.99:
        gs_style = 1
    elif gs_style < 0.01:
        gs_style = 0.01
    tan_cutoff = (1 - gs_style) * acc_range + min_tan

    # Search_from = int(tmp_min_next * 10)
    search_from = int((tmp_min_next - tmp_min) * 10) + 1

    i_cut = len(tan) - 1
    while tan[i_cut] < tan_cutoff and i_cut >= search_from:
        i_cut -= 1

    gear_cut = tmp_min + i_cut / 10 + 0.1

    return gear_cut


def gear_points_from_tan(Tans, gs_style, Start, Stop):
    """

    Get the gear cuts based on gear shifting style and tangent values.

    :param Tans: tangent values per gear.
    :param gs_style: Gear shifting style
    :param Start: Start speed per gear curve.
    :param Stop: Stop speed per gear curve.
    :return:
    """
    n_gears = len(Tans)
    gs_cut = [gs_style for i in range(n_gears)]

    gs = []

    for i in range(n_gears - 1):
        tmp_min = Start[i]
        tmp_max = Stop[i]
        tan = Tans[i]
        tmp_min_next = Start[i + 1]
        cutoff_s = find_gs_cut_tans(tmp_min, tmp_max, tan, tmp_min_next, gs_cut[i])

        gs.append(cutoff_s)

    return gs


def gear_for_speed_profiles(gs, curr_speed, current_gear, gear_cnt, clutch_duration=5):
    """
    Return the gear that must be used and the clutch condition

    :param gs:
    :param curr_speed:
    :param current_gear:
    :param gear_cnt:
    :param clutch_duration: in sim step
    :return:
    """

    # Model buffer for up shifting and down shifting.
    upshift_offs = 0.0
    downshift_off = 0.1

    gear_limits = [0]
    gear_limits.extend(gs)
    gear_limits.append(200)

    if gear_limits[current_gear - 1] - gear_limits[current_gear - 1] * downshift_off <= curr_speed < gear_limits[
        current_gear] + gear_limits[current_gear] * upshift_offs:
        if gear_cnt == 0:
            return current_gear, gear_cnt
        else:
            gear_cnt -= 1
            return current_gear, gear_cnt
    else:
        iter = 1
        gear_search = 1
        while iter == 1:
            if gear_limits[gear_search - 1] <= curr_speed < gear_limits[gear_search]:
                gear_cnt = clutch_duration  # in simulation steps for 0.5 second
                current_gear = gear_search
                iter = 0
            else:
                gear_search += 1
        return current_gear, gear_cnt


def gear_linear(speed_per_gear, gs_style):
    """
    Return the gear limits based on gs_style, using linear gear swifting
    strategy

    :param speed_per_gear:
    :param gs_style:
    :return:
    """
    n_gears = len(speed_per_gear)

    gs_style = min(gs_style, 1)
    gs_style = max(gs_style, 0)

    gs = []

    for gear in range(n_gears - 1):
        speed_by_gs = speed_per_gear[gear][-1] * gs_style + \
                      speed_per_gear[gear][0] * (1 - gs_style)
        speed_for_continuity = speed_per_gear[gear + 1][0]
        cutoff_s = max(speed_by_gs, speed_for_continuity)

        gs.append(cutoff_s)

    return gs


def create_clutch_list(gear_list, clutch_duration):
    """

    Returns a list with the vehicle clutch position (1 or 0) for each simulation step

    :param gear_list: A list with the vehicle gear on each simulation step
    :param clutch_duration: The duration of the clutch, in sim steps
    :return:
    """

    l1 = len(gear_list) + clutch_duration
    clutch_list = [0]*l1

    for i,gear in enumerate(gear_list[1:]):
        if gear != gear_list[i]:
            clutch_list[i+1:i+clutch_duration] = [1] * clutch_duration

    if clutch_duration != 0:
        clutch_list = clutch_list[:-clutch_duration]

    return clutch_list


import schedula as sh
import numpy.testing as nt


def _assert(v, r, name):
    if isinstance(v, str):
        nt.assert_equal(v, r, 'Value %s is not equal!' % name)
    elif isinstance(v, dict):
        for k, v in v.items():
            _assert(v, r[k], '{}/{}'.format(name, k))
    elif isinstance(v, list):
        for i, (v, r) in enumerate(zip(v, r)):
            _assert(v, r, '{}/{}'.format(name, i))
    elif callable(v):
        pass
    else:
        nt.assert_almost_equal(v, r, err_msg='Value %s is not equal!' % name)


def _check(dsp, data, outputs):
    inputs = sh.selector(set(data) - set(outputs), data)
    res = dsp(inputs, outputs)
    for d in res.workflow.nodes.values():
        if 'solution' not in d:
            continue
        s = d['solution']
        nt.assert_equal(
            bool(s._errors), False, "Found errors in {}".format(set(s._errors))
        )
    nt.assert_equal(
        set(outputs).issubset(res), True,
        "Missing outputs {}".format(set(outputs) - set(res))
    )
    for k in outputs:
        _assert(data[k], res[k], k)



from os import path as osp, chdir
import matplotlib.pyplot as plt
import numpy as np
from co2mpas_driver import dsp as driver

my_dir = osp.dirname(osp.abspath(__file__))
chdir(my_dir)


def simple_run():
    """
    This sample file plots the speed/acceleration curves of the vehicle and
    regression results of 2nd degree polynomial based on gear shifting style.

    :return:
    """
    # A sample car id from the database
    car_id = 27748

    # The gear shifting style as described in the TRR paper.
    gs_style = 1

    # 2nd degree polynomial
    degree = 2

    # Core model, this will select and execute the proper functions for the given inputs and returns the output
    # You can also pass your vehicle's database path db_path='path to vehicle db'
    sol = driver(dict(vehicle_id=car_id, inputs=dict(inputs=dict(
        gear_shifting_style=gs_style, degree=degree, use_linear_gs=False,
        use_cubic=True))))['outputs']
    # Plots workflow of the core model, this will automatically open an internet browser and shows the work flow
    # of the core model. you can click all the rectangular boxes to see in detail sub models like load, model,
    # write and plot.
    # you can uncomment to plot the work flow of the model
    # driver.plot(1)  # Note: run your IDE as Administrator if file permission error.

    # gear cuts based on linear gear shifting style
    gs = sol['gs']

    # coefficients of speed acceleration relation for each gear
    # calculated using 2nd degree polynomial
    coefs_per_gear = sol['coefs_per_gear']
    speed_per_gear = sol['speed_per_gear']
    acc_per_gear = sol['acc_per_gear']

    degree = len(coefs_per_gear[0]) - 1
    vars = np.arange(degree, -1, -1)

    fig = plt.figure('speed acceleration regression results of degree = ' + str(degree))
    for gear in gs:
        plt.plot([gear, gear], [0, 5])

    for gear, (speeds, acceleration, fit_coef) in enumerate(zip(speed_per_gear, acc_per_gear, coefs_per_gear)):
        plt.plot(speeds, acceleration, label=f'gear {gear}')
        x_new = np.arange(speeds[0], speeds[-1], 0.1)
        a_new = np.array([np.dot(fit_coef, np.power(i, vars)) for i in x_new])
        plt.plot(x_new, a_new, label=f'{degree} degree reg. gear {gear}')
    plt.legend()
    plt.text(46, 1.8, f'Simulation car:{car_id}, GS:{gs_style}')
    fig.suptitle('Speed acceleration regression results of degree = ' + str(degree), x=0.54, y=1, fontsize=12)
    plt.xlabel('Speed (m/s)', fontsize=12)
    plt.ylabel('Acceleration (m/s2)', fontsize=12)
    plt.show()


if __name__ == '__main__':
    simple_run()


from os import path as osp, chdir
from co2mpas_driver.common import simulation_part as sp
import numpy as np
import matplotlib.pyplot as plt
from co2mpas_driver.common import curve_functions as mf
from co2mpas_driver.common import reading_n_organizing as rno
from co2mpas_driver.common import gear_functions as fg

my_dir = osp.dirname(osp.abspath(__file__))
chdir(my_dir)


def simple_run():
    """
    This example performs a simulation with MFC model from a starting speed to a desired speed.

    :return:
    """
    # Vehicle databased based on the Euro Car Segment classification
    db_path = osp.abspath(osp.join(osp.dirname(my_dir + '/../'), 'db', 'EuroSegmentCar'))

    # A sample car id from the database
    car_id = 35135

    # The gear shifting style as described in the TRR paper.
    gs_style = 0.8

    # 4th degree polynomial
    degree = 2

    # The desired speed
    vdes = 124/3.6

    # Current speed
    v_start = 0

    # The simulation step in seconds
    sim_step = 0.1

    # The driving style as described in the TRR paper.
    driver_style = 0.7

    # Duration of the simulation in seconds.
    duration = 100

    # sample time series
    times = np.arange(10, duration + sim_step, sim_step)

    # load vehicle database
    db = rno.load_db_to_dictionary(db_path)

    # The vehicle specs as returned from the database
    selected_car = rno.get_vehicle_from_db(db, car_id)

    '''
    The final acceleration curvers (Curves), the engine acceleration potential 
    curves (cs_acc_per_gear), before the calculation of the resistances and the 
    limitation due to max possible acceleration (friction) .
    '''
    # Curves, Curves_dec, StartStop, gs = mf.get_ev_curve_main(selected_car)

    Curves, cs_acc_per_gear, StartStop, gs = mf.gear_curves_n_gs_from_poly(selected_car, gs_style, degree)

    # Lists to gather simulation data
    Speeds = [v_start]
    Acceleration = [0]

    # Initialize speed and gear
    speed = v_start

    # Returns the gear that must be used and the clutch condition
    gear, gear_count = fg.gear_for_speed_profiles(gs, speed, 0, 0)
    gear_count = 0

    # Core loop
    for t in times:
        speed, acceleration, gear, gear_count = sp.simulation_step_function(
            selected_car, speed, gear, gear_count, gs, Curves,
            vdes, driver_style, sim_step)
        # Gather data
        Speeds.append(speed)
        Acceleration.append(acceleration)

    '''Plot'''
    fig2 = plt.figure()
    plt.plot(times, Speeds[1:])
    plt.text(30, 22, f'Simulation car:{car_id}, DS:{driver_style}, GS:{gs_style}')
    fig2.suptitle('Speed over time', x=0.54, y=1, fontsize=12)
    plt.xlabel('Time (sec)', fontsize=12)
    plt.ylabel('Speed (m/s)', fontsize=12)
    plt.grid()

    fig3 = plt.figure('Acceleration over Speed')
    plt.plot(Speeds[1:], Acceleration[1:], 'k', label='simulation')
    fig3.suptitle('Acceleration over Speed', x=0.54, y=1, fontsize=12)
    plt.xlabel('Speed (m/s)', fontsize=12)
    plt.ylabel('Acceleration (m/s2)', fontsize=12)

    fig4 = plt.figure('Acceleration over time')
    plt.plot(times, Acceleration[1:])
    plt.text(23, 2.2, f'Simulation car:{car_id}, DS:{driver_style}, GS:{gs_style}')
    fig4.suptitle('Acceleration over time', x=0.54, y=1, fontsize=12)
    plt.xlabel('Time (sec)', fontsize=12)
    plt.ylabel('Acceleration (m/s2)', fontsize=12)
    plt.grid()

    fig5 = plt.figure('Acceleration over Speed')
    for gear, gear_curve in enumerate(Curves, start=1):
        sp_bins = np.arange(StartStop[0][gear - 1], StartStop[1][gear - 1] + 0.1, 0.1)
        accelerations = gear_curve(sp_bins)
        plt.plot(sp_bins, accelerations, label=f'gear {gear}')
    plt.legend()
    plt.text(23, 2.5, f'Simulation car:{car_id}, DS:{driver_style}, GS:{gs_style}')
    fig5.suptitle('Acceleration over Speed', x=0.54, y=1, fontsize=12)
    plt.xlabel('Speed (m/s)', fontsize=12)
    plt.ylabel('Acceleration (m/s2)', fontsize=12)
    plt.grid()

    plt.show()


if __name__ == '__main__':
    simple_run()


'''Define veh_specs class'''


class veh_specs(object):

    # The class "constructor"
    def __init__(self, my_car, **kwargs):
        '''
        kwargs can be:
        lco = True          # Light co2mpas is to be used, so the relevant parameters must be imported
        electric = True     # The vehicle is an EV

        :param my_car:
        :param kwargs:
        '''

        if 'electric' not in kwargs or not kwargs['electric']:
            self.engine_max_power = float(my_car["Fuel Engine-Max power"])  # kW
            self.engine_max_speed_at_max_power = float(my_car["Fuel Engine-Max power RPM"])  # rpm
            gr_str = my_car["Transmission  / Gear ratio-Gear Box Ratios"]  # [3.33, 1.95, 1.29, 0.98, 0.76]
            if gr_str[0] == '[':
                gr_str = gr_str[1:-1]
            gr_str = gr_str.split('-')
            gr = []
            for i in range(len(gr_str)):
                gr.append(float(gr_str[i]))
            self.gr = gr

            self.fuel_type = my_car["Drive-Fuel"]
            # What happens with the other types of fuels?
            if self.fuel_type == 'petrol':
                self.ignition_type = 'positive'
            elif self.fuel_type == 'diesel':
                self.ignition_type = 'compression'
            else:
                print("neither petrol nor diesel, I exit!")
                exit()

            self.tire_radius = (24.5 / 2 * 2.54) / 100  # meters
            self.driveline_slippage = 0
            if my_car["General Specifications-Transmission"] == 'automatic':
                self.driveline_efficiency = 0.90
                self.transmission = 'automatic'
            else:
                self.driveline_efficiency = 0.93
                self.transmission = 'manual'
            self.final_drive = float(my_car["Transmission  / Gear ratio-Final drive"])
            self.veh_mass = float(my_car["Weights-Empty mass"])
            self.top_speed = int(float(my_car["Performance-Top speed"]) / 3.6)
            self.type_of_car = my_car["General Specifications-Carbody"].strip()
            self.car_width = float(my_car["Exterior sizes-Width"])
            self.car_height = float(my_car["Exterior sizes-Height"])
            # self.kerb_weight = float(my_car["Weights-Unladen mass"])
            self.wheelbase = float(my_car["Exterior sizes-Wheelbase"])
            car_type = my_car["Drive-Wheel drive"]
            if car_type == 'front':
                self.car_type = 2
            elif car_type == 'rear':
                self.car_type = 4
            else:
                self.car_type = 6

                # To check it again in the future
            if self.ignition_type == "positive":
                self.idle_engine_speed = (750, 50)
            else:
                self.idle_engine_speed = (850, 50)

            self.veh_max_speed = round(int(float(my_car["Performance-Top speed"]) / 3.6), 2)  # m/s

            if 'lco' in kwargs:
                if kwargs['lco']:
                    '''Used for Light Co2mpass'''
                    self.r_dynamic = float(my_car["Chassis-Rolling Radius Dynamic"]) / 1000
                    self.engine_max_torque = float(my_car["Fuel Engine-Max torque"])
                    self.fuel_engine_stroke = float(my_car["Fuel Engine-Stroke"])
                    self.max_power = float(my_car["Drive-Total max power"])
                    self.fuel_turbo = my_car["Fuel Engine-Turbo"]
                    self.fuel_eng_capacity = float(my_car["Fuel Engine-Capacity"])
                    self.gearbox_type = str(my_car["General Specifications-Transmission"])

        else:
            if kwargs['electric']:
                self.fuel_type = my_car["Drive-Fuel"]
                if self.fuel_type != 'electricity':
                    raise ("NOT ELECTRIC")
                self.ignition_type = 'electricity'

                self.engine_max_power = float(my_car["Electric Engine-Total max power"])  # kW
                self.motor_max_torque = float(my_car["Electric Engine-Max torque"])  # Nm
                self.gr = 1
                self.tire_radius = float(my_car["Chassis-Rolling Radius Static"]) / 1000  # meters
                self.driveline_slippage = 0
                if my_car["General Specifications-Transmission"] == 'single-speed fixed gear':
                    # self.driveline_efficiency = 0.90
                    self.driveline_efficiency = 0.93
                else:
                    self.driveline_efficiency = 0.93
                self.final_drive = float(my_car["Transmission  / Gear ratio-Final drive"])
                self.veh_mass = float(my_car["Weights-Empty mass"])
                self.veh_max_speed = round(int(float(my_car["Performance-Top speed"]) / 3.6), 2)  # m/s
                self.type_of_car = my_car["General Specifications-Carbody"].strip()
                self.car_width = float(my_car["Exterior sizes-Width"])
                self.car_height = float(my_car["Exterior sizes-Height"])
                # self.kerb_weight = float(my_car["Weights-Unladen mass"])
                self.car_type = my_car["Drive-Wheel drive"]
                if self.car_type == 'front':
                    self.car_type = 2
                elif self.car_type == 'rear':
                    self.car_type = 4
                else:
                    self.car_type = 6


class hardcoded_params(object):
    '''
    The class is used for some params that are hard coded
    '''

    # The class "constructor"
    def __init__(self):
        # PARAMETERS HARDCODED - TO BE LOADED FROM OUTSIDE
        self.final_drive_efficiency = 0.98
        self.thres = 1100
        self.min_engine_on_speed = 600

        self.params_gearbox_losses = {
            'Automatic': {
                'gbp00': {'m': -0.0043233434399999994,
                          'q': -0.29823614099999995},
                'gbp10': {'m': -2.4525999999999996e-06,
                          'q': -0.0001547871},
                'gbp01': {'q': 0.9793688500000001},
            },
            'Manual': {
                'gbp00': {'m': -0.0043233434399999994,
                          'q': -0.29823614099999995},
                'gbp10': {'m': -2.4525999999999996e-06,
                          'q': -5.15957e-05},
                'gbp01': {'q': 0.9895177500000001},
            }
        }
        # FIX It is petrol not gasoline in the car db.
        #: Fuel density [g/l].
        self.FUEL_DENSITY = {
            'petrol': 745.0,
            'diesel': 832.0,
            'LPG': 43200.0 / 46000.0 * 745.0,  # Gasoline equivalent.
            'NG': 43200.0 / 45100.0 * 745.0,  # Gasoline equivalent.
            'ethanol': 794.0,
            'biodiesel': 890.0,
        }

        self.LHV = {
            'petrol': 43200.0,
            'diesel': 43100.0,
            'LPG': 46000.0,
            'NG': 45100.0,
            'ethanol': 26800.0,
            'biodiesel': 37900.0,
        }

        self.idle_engine_speed_median = {
            'petrol': self.min_engine_on_speed + 300,
            'diesel': self.min_engine_on_speed + 200
        }



from os import path as osp, chdir
import numpy as np
import matplotlib.pyplot as plt
from co2mpas_driver import dsp as driver

my_dir = osp.dirname(osp.abspath(__file__))
chdir(my_dir)


def simple_run():
    """
     This example computes and plots the fuel consumption in grams for a
     simulated trajectory.

    """
    # A sample car id from the database = {
    car_id_db = {
        'fuel_engine': [35135, 39393, 27748, 8188, 40516, 35452, 40225, 7897, 7972, 41388, 5766, 9645, 9639, 5798, 8280,
                        34271, 34265, 6378, 39723, 34092, 2592, 5635, 5630, 7661, 7683, 8709, 9769, 1872, 10328, 35476,
                        41989, 26799, 26851, 27189, 23801, 3079, 36525, 47766, 6386, 33958, 33986, 5725, 5718, 36591,
                        4350, 39396, 40595, 5909, 5897, 5928, 5915, 40130, 42363, 34760, 34766, 1835, 36101, 42886,
                        1431, 46547, 44799, 41045, 39820, 34183, 34186, 20612, 20605, 1324, 9882, 4957, 5595, 18831,
                        18833, 9575, 5380, 9936, 7995, 6331, 18173, 34286, 34279, 20706, 34058, 34057, 24268, 19028,
                        19058, 7979, 22591, 34202, 40170, 44599, 5358, 5338, 34015, 9872, 9856, 6446, 8866, 9001, 9551,
                        6222],
        'electric_engine': [47844]
    }
    car_id = 39393

    # The desired speed
    vdes = 124 / 3.6

    # Current speed
    v_start = 0

    # The simulation step in seconds
    sim_step = 0.1

    # The driving style as described in the TRR paper.
    driver_style = 0.8

    # Duration of the simulation in seconds.
    duration = 100

    # sample time series
    times = np.arange(0, duration + sim_step, sim_step)

    # core model, this will select and execute the proper functions for the given inputs and returns the output
    # You can also pass vehicle database path db_path='path to vehicle db'
    if car_id in car_id_db['electric_engine']:
        sol = driver(dict(vehicle_id=47844, inputs=dict(inputs=dict(
            desired_velocity=vdes, starting_velocity=v_start, driver_style=driver_style,
            sim_start=0, sim_step=sim_step, duration=duration, use_linear_gs=True,
            use_cubic=False))))['outputs']
    else:
        # The gear shifting style as described in the TRR paper.
        gs_style = 0.8
        sol = driver(dict(vehicle_id=car_id, inputs=dict(inputs=dict(
            gear_shifting_style=gs_style, desired_velocity=vdes,
            starting_velocity=v_start, driver_style=driver_style,
            sim_start=0, sim_step=sim_step, duration=duration, degree=4, use_linear_gs=True,
            use_cubic=False))))['outputs']
    # Plots workflow of the core model, this will automatically open an internet browser and show the work flow
    # of the core model. you can click all the rectangular boxes to see in detail sub models like load, model,
    # write and plot.
    # you can uncomment to plot the work flow of the model
    # driver.plot(1)  # Note: run your IDE as Administrator if file permission error.

    fig = plt.figure('Time-Speed')
    plt.plot(times, sol['velocities'][1:], label=f'Simulation car:{car_id}, DS: {driver_style}')
    plt.plot(times, np.ones(len(times)) * vdes, label=f'Desired speed (m/s)')
    plt.legend()
    fig.suptitle('Speed over time', x=0.54, y=1, fontsize=12)
    plt.xlabel('Time (s)', fontsize=14)
    plt.ylabel('Speed (m/s)', fontsize=12)
    plt.grid()

    fig = plt.figure('Time-Acceleration')
    plt.plot(times, sol['accelerations'][1:], label=f'Simulation car:{car_id}, DS: {driver_style}')
    plt.legend()
    fig.suptitle('Acceleration over time', x=0.54, y=1, fontsize=12)
    plt.xlabel('Time (s)', fontsize=14)
    plt.ylabel('Acceleration (m/s2)', fontsize=12)
    plt.grid()

    fig = plt.figure('Speed-Acceleration')
    plt.plot(sol['velocities'][1:], sol['accelerations'][1:], label=f'Simulation car:{car_id}, DS: {driver_style}')
    plt.legend()
    fig.suptitle('Acceleration over speed', x=0.54, y=1, fontsize=12)
    plt.xlabel('Speed (m/s)', fontsize=14)
    plt.ylabel('Acceleration (m/s2)', fontsize=12)
    plt.grid()

    if car_id not in car_id_db['electric_engine']:
        fig = plt.figure('Time-Fuel consumption')
        plt.plot(times, sol['fp'], label=f'Simulation car:{car_id}, DS: {driver_style}')
        plt.legend()
        fig.suptitle('Fuel Consumption over time', x=0.54, y=1, fontsize=12)
        plt.xlabel('Time (s)', fontsize=14)
        plt.ylabel('Fuel Consumption (g)', fontsize=12)
        plt.grid()

    plt.show()


if __name__ == '__main__':
    simple_run()



from os import path as osp, chdir
import pandas as pd
from plotly import graph_objs as go
from co2mpas_driver.common import reading_n_organizing as rno

my_dir = osp.dirname(osp.abspath(__file__))
chdir(my_dir)


def simple_run():
    """
    This example illustrates the specifications of vehicles in the database.

    :return:
    """
    # Vehicle database based on the Euro Car Segment classification
    db_path = osp.abspath(osp.join(osp.dirname(my_dir + '/../'), 'db', 'EuroSegmentCar_cleaned'))
    
    # load vehicle database
    db = rno.load_db_to_dictionary(db_path)
    df = pd.DataFrame.from_dict(db, orient='index')
    df = df.reset_index(drop=True)
    df['vehicle-class'] = pd.Categorical(df['vehicle-class'])
    df['vehicle-class'] = df['vehicle-class'].cat.codes

    # fig = px.parallel_coordinates(df_veh, color="Drive-Total max power",
    #                             dimensions=['vehicle-class', 'Drive-Total max power', 'Exterior sizes-Length', 'Performance-Acceleration 0-100 km/h',
    #                                         'Performance-Top speed', 'Weights-Empty mass', 'Class-Euro standard car segments'],
    #                             color_continuous_scale=px.colors.diverging.Tealrose,
    #                             color_continuous_midpoint=2)

    fig = go.Figure(data=go.Parcoords(
            line=dict(color=df['Drive-Total max power'].astype(float),
                    colorscale=[[0, 'purple'], [0.5, 'lightseagreen'], [1, 'gold']]),
            dimensions=[
                dict(
                    label='Max power (kW)', values=df['Drive-Total max power'].astype(float)),
                dict(
                    label='Max torque (Nm)', values=df['Drive-Total max torque'].astype(float)),
                dict(
                    label='Max speed (m/s)', values=df['Performance-Top speed'].astype(float)),
                dict(
                    label='Veh mass (kg)', values=df['Weights-Empty mass'].astype(float)),
                dict(
                    label='Veh length (m)', values=df['Exterior sizes-Length'].astype(float)),
                dict(
                    label='0-100km/h time (s)', values=df['Performance-Acceleration 0-100 km/h'].astype(float)),
                dict(tickvals=list(range(9)), ticktext=['A', 'B', 'C', 'D', 'E', 'F', 'J', 'M', 'S'], label='Veh class',
                     values=df['vehicle-class'])
            ]
        )
    )

    fig.update_layout(
        plot_bgcolor='white',
        paper_bgcolor='white'
    )
    fig.show()


if __name__ == '__main__':
    simple_run()


# -*- coding: utf-8 -*-
#
# Copyright 2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions to process a CO2MPAS input file.
"""

# Computation of the MFC vehicle acceleration - speed curve.
# coding=utf-8
import functools
import numpy as np
from scipy.interpolate import CubicSpline
from co2mpas_driver.common import defaults as defaults


def get_full_load(ignition_type):
    """
    Get vehicle full load curve.

    :param ignition_type:
        Engine ignition type (positive or compression).
    :type ignition_type: str

    :return:
        Vehicle normalized full load curve.
    :rtype: scipy.interpolate.InterpolatedUnivariateSpline
    """

    xp, fp = defaults.dfl.functions.get_full_load.FULL_LOAD[ignition_type]
    func = functools.partial(
        np.interp, xp=xp, fp=fp, left=fp[0], right=fp[-1]
    )

    return func


def calculate_full_load_speeds_and_powers(
        full_load_curve, engine_max_power, engine_max_speed_at_max_power,
        idle_engine_speed):
    """
    Calculates the full load speeds and powers [RPM, kW].

    :param full_load_curve:
        Vehicle normalized full load curve.
    :type full_load_curve: scipy.interpolate.InterpolatedUnivariateSpline

    :param engine_max_power:
        Engine nominal power [kW].
    :type engine_max_power: float

    :param engine_max_speed_at_max_power:
        Engine nominal speed at engine nominal power [RPM].
    :type engine_max_speed_at_max_power: float

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :return:
         T1 map speed [RPM] and power [kW] vectors.
    :rtype: (numpy.array, numpy.array)
    """

    n_norm = np.arange(0.0, 1.21, 0.1)
    full_load_powers = full_load_curve(n_norm) * engine_max_power
    idle = idle_engine_speed[0]
    full_load_speeds = n_norm * (engine_max_speed_at_max_power - idle) + idle

    return full_load_speeds, full_load_powers


def calculate_full_load_torques(full_load_speeds, full_load_powers):
    """
    Calculate Full load curves of speed and torque.

    :param full_load_speeds:
        Full load speeds.
    :type full_load_powers: str

    :param full_load_powers:
        Engine nominal power [kW].
    :type full_load_speeds: float

    :return: full_load_torques
        Full load torques.
    :rtype:
    """
    full_load_torques = full_load_powers * 1000 * (
            full_load_speeds / 60 * 2 * np.pi) ** -1

    return full_load_torques


# The maximum acceleration that the vehicle can have on the road
def Armax(car_type, veh_mass, engine_max_power, road_type=1):
    """
    Calculating the maximum acceleration possible for the vehicle object my_car,
    under road_type conditions.

    :param car_type:
        Car type.
    :type car_type: int

    :param veh_mass:
        Vehicle mass.
    :type veh_mass: float

    :param engine_max_power:
        Maximum engine power.
    :type engine_max_power:

    :param road_type:
        Road type(1: normal, 2: wet, 3: icy)
    :type road_type: int

    :return:
        Maximum possible acceleration of the vehicle.
    :rtype:
    """
    if car_type == 2:  # forward-wheel drive vehicles
        f_mass = 0.6 * veh_mass
    elif car_type == 4:  # rear-wheel drive vehicles
        f_mass = 0.45 * veh_mass
    else:  # all-wheel drive vehicles, 4x4
        f_mass = 1 * veh_mass

    if road_type == 1:
        mh_base = 0.75  # for normal road
    elif road_type == 2:
        mh_base = 0.25  # for wet road
    else:
        mh_base = 0.1  # for icy road
    # Optimal values:
    # 0.8 dry, 0.6 light rain, 0.4 heavy rain, 0.1 icy
    alpha = 43.398
    beta = 5.1549
    mh = mh_base * (alpha * np.log(engine_max_power) + beta) / 190

    # * cos(f) for the gradient of the road. Here we consider as 0
    # maximum achieved force by the vehicles in certain conditions.
    f_max = f_mass * 9.8066 * mh

    return f_max / veh_mass


# Calculates a spline with the resistances
def veh_resistances(f0, f1, f2, sp_bins, total_mass):
    """
    Calculate the resistances that a vehicle faces, per speed.

    :param f0:
        Tire rolling resistance.
    :type f0: float

    :param f1:
        Partly tire rolling resistance & partly drivetrain losses.
    :type f1: float

    :param f2:
        Aerodynamic component (proportional to the square of the vehicles
        velocity)
    :type f2: float

    :param sp_bins:
        Speed bins.
    :type sp_bins: list[float]

    :param total_mass:
        Total mass.
    :type total_mass: float

    :return: resistance_spline_curve, resistance_spline_curve_f
        Resistance forces being applied per speed.
    :rtype: scipy.interpolate._cubic.CubicSpline, scipy.interpolate._cubic.CubicSpline
    """
    sp_bins = list(sp_bins)
    resistance_force = []
    for i in range(len(sp_bins)):
        resistance_force.append(f0 + f1 * sp_bins[i] * 3.6 + f2 * pow(sp_bins[i] * 3.6, 2))
        # Facc = Fmax @ wheel - f0 * cos(a) - f1 * v - f2 * v2 - m * g * sin(a)

    approximate_mass = int(total_mass)
    acc_resistance = [x / approximate_mass for x in resistance_force]
    a = int(np.floor(sp_bins[0]))
    b = int(np.floor(sp_bins[-1]))
    resistance_spline_curve = CubicSpline(
        [k for k in range(a - 10, a)] + sp_bins + [k for k in range(b + 1, b + 11)], \
        [acc_resistance[0]] * 10 + acc_resistance + [acc_resistance[-1]] * 10)
    resistance_spline_curve_f = CubicSpline(
        [k for k in range(a - 10, a)] + sp_bins + [k for k in range(b + 1, b + 11)],
        [resistance_force[0]] * 10 + resistance_force + [resistance_force[-1]] * 10)

    return resistance_spline_curve, resistance_spline_curve_f


def estimate_f_coefficients(veh_mass, type_of_car, car_width, car_height,
                            passengers=0):
    """
    Estimate f0, f1, f2 coefficients of resistance.

    :param veh_mass:
        Vehicle mass.
    :type veh_mass: float

    :param type_of_car:
        Type of car.
    :type type_of_car: str

    :param car_width:
        Car width.
    :type car_width: float

    :param car_height:
        Car height.
    :type car_height: float

    :param passengers:
        number of passengers.
    :type passengers: int

    :return:
        Coefficients of resistance force.
    :rtype: float, float, float
    """

    d = {}
    # Fill in the entries one by one
    d["cabriolet"] = 0.28
    d["sedan"] = 0.27
    d["hatchback"] = 0.3
    d["stationwagon"] = 0.28
    d["suv/crossover"] = 0.35
    d["mpv"] = 0.3
    d["coupé"] = 0.27
    d["pick-up"] = 0.4

    rolling_res_coef = 0.009  # Constant for the moment
    theor_aero_coeff = d[type_of_car]

    operating_mass = veh_mass + 100 + 75 * passengers
    f0 = (operating_mass + 100) * rolling_res_coef * 9.81
    f2 = 0.5 * 1.2 * (
            0.84 * car_width * car_height * theor_aero_coeff) / pow(
        3.6, 2)
    f1 = -71.735 * f2 + 2.7609

    return f0, f1, f2



import matplotlib
from co2mpas_driver import dsp as driver
import matplotlib.pyplot as plt
import numpy as np
import math as m

matplotlib.use('qt4agg')


def idm_calc_vel(af_mfc, fol_v, lead_v, vn, dist, bn, an, sim_step, **kwargs):
    if 'ds' in kwargs:  # intervehicle spacing at a stop
        ds = kwargs['ds']
    else:
        ds = 2
    if 'th' in kwargs:  # the desired time headway to the vehicle in front
        th = kwargs['th']
    else:
        th = 1.5
    if 'delta' in kwargs:  # the desired time headway to the vehicle in front
        delta = kwargs['delta']
    else:
        delta = 4
    bn_idm = abs(bn)  # braking deceleration  b - a positive number
    return max(0, fol_v + (af_mfc + \
                           -an * (pow(
                max(0, (ds + fol_v * th) / dist + (fol_v * (fol_v - lead_v)) / (2 * m.sqrt(an * bn_idm) * dist)),
                2))) * sim_step)


# IDM params
amax = 3
bn = -2
delta = 4
ds = 2
th = 1.5

np.random.seed(123)

# veh_ids = [27748, 8188]
# veh_ids = [39393, 8188, 40516, 35452, 40225, 7897, 7972, 41388, 5766,
#            9645, 9639, 5798, 8280, 34271, 34265, 6378, 39723, 34092, 2592,
#            5635, 5630, 7661, 7683]
veh_pool = [39393, 8188, 40516, 35452, 40225, 7897, 7972, 41388, 5766,
            9645, 9639, 5798, 8280, 34271, 34265, 6378, 39723, 34092, 2592,
            5635, 5630, 7661, 7683]
num_vehs = 1
veh_ids = [np.random.choice(veh_pool) for i in range(num_vehs)]

# v_des = 100 / 3.6
v_start = 0
dt = 0.1

sim_duration = 230.1  # sec

times = np.arange(0, sim_duration, dt)

vehicles = [driver(dict(vehicle_id=i, inputs=dict(inputs=dict(
    gear_shifting_style=1, driver_style=1,
    use_linear_gs=True,
    use_cubic=False))))['outputs'] for i in
            veh_ids]
models = [veh['driver_simulation_model'] for veh in vehicles]
dcurves = [veh['driver_simulation_model'] for veh in vehicles]

dist = 30  # meters

test_leader_sp_profile = [20 / 3.6] * int(10 * sim_duration / 6) + [50 / 3.6] * int(10 * sim_duration / 6) + [
    120 / 3.6] * int(10 * sim_duration / 6) + [80 / 3.6] * int(10 * sim_duration / 6) + [
                             60 / 3.6] * int(10 * sim_duration / 6) + [100 / 3.6] * int(10 * sim_duration / 6)
leader_dt = np.cumsum([i * dt for i in test_leader_sp_profile]) + dist * (len(models) + 1)

res = {}
cnt = 0

v_des_p = []
my_veh = models[0]
my_veh.reset(v_start)
res[my_veh] = {'accel': [0], 'speed': [v_start], 'pos': [0]}
for idx, myt in enumerate(times):
        dd = 700
        if 0 <= res[my_veh]['pos'][idx] <= 0.3 * dd:
            v_des = 20
        elif 0.3 * dd < res[my_veh]['pos'][idx] <= 1 * dd:
            v_des = 30
        elif 1 * dd < res[my_veh]['pos'][idx] <= 1.5 * dd:
            v_des = 5
        elif 1.5 * dd < res[my_veh]['pos'][idx] <= 2 * dd:
            v_des = 25
        elif 2 * dd < res[my_veh]['pos'][idx] <= 4.5 * dd:
            v_des = 45
        elif 4.5 * dd < res[my_veh]['pos'][idx] <= 6.5 * dd:
            v_des = 15
        elif 6.5 * dd < res[my_veh]['pos'][idx] <= 7.5 * dd:
            v_des = 0

        v_des_p.append(v_des)
        # Compute MFC gear, next_velocity, acceleration - The positing is not needed
        gear, gear_count, next_velocity, acc = my_veh(dt, v_des)

        final_acc = acc

        # So the next velocity will be based on the cf acceleration (not the MFC one)
        next_velocity = res[my_veh]['speed'][-1] + final_acc * dt

        ####
        if next_velocity < 0:
            next_velocity = 0
        if v_des - 0.001 < next_velocity < v_des + 0.001:
            next_velocity = v_des
        # if next_velocity > v_des:
        #     next_velocity = v_des
        ####

        # I save my results. Again here positing is of no use
        res[my_veh]['accel'].append(final_acc)
        res[my_veh]['speed'].append(next_velocity)
        res[my_veh]['pos'].append(res[my_veh]['pos'][idx]+ (res[my_veh]['speed'][idx]+res[my_veh]['speed'][-1])*0.5*dt)

        # if res[models[vidx]]['pos'][idx - 1] + next_velocity * (dt) + 0.5 * acc * dt ** 2 <0:
        #     print('check')
        # dtf = res[models[vidx]]['pos'][idx - 1] + next_velocity * (dt) + 0.5 * acc * dt ** 2
        # res[models[vidx]]['pos'].append(dtf)

        # update the vehicle for the next MFC
        gear, gear_count = my_veh.update(next_velocity, gear, gear_count, dt)
        cnt += 1
        if cnt == 1651:
            print('check')
        print(cnt)

fig = plt.figure('Trajectory (distance-speed)')
ax = fig.add_subplot(111)
ax.plot(res[my_veh]['pos'][1:], v_des_p, label='Reference speed')
ax.plot(res[my_veh]['pos'][1:], res[my_veh]['speed'][1:], label='Reference speed')
plt.show()



import math
import numpy


def calculate_wheel_power(velocities, acc, road_loads, veh_mass, slope):
    """
    Calculates the wheel power [kW].

    :param velocities:
        Velocity [km/h].
    :type velocities: numpy.array | float

    :param acc:
        Acceleration [m/s2].
    :type acc: numpy.array | float

    :param road_loads:
        Cycle road loads [N, N/(km/h), N/(km/h)^2].
    :type road_loads: list, tuple

    :param veh_mass:
        Vehicle mass [kg].
    :type veh_mass: float

    :param slope:
        Slope.
    :type slope: float

    :return:
        Power at wheels [kW].
    :rtype: numpy.array | float

    ;param slope:
        Slope in tangent
    """

    f0, f1, f2 = road_loads

    # quadratic_term = f0 * math.cos(math.atan(slope)) + (f1 + f2 * velocities) * velocities
    #
    # vel = velocities / 3600

    quadratic_term = f0 * math.cos(math.atan(slope)) + (
            f1 + f2 * velocities) * velocities + 1.03 * veh_mass * acc + veh_mass * math.sin(
        math.atan(slope)) * 9.81

    res = float(quadratic_term * velocities) / 3600

    return res


def calculate_min_wheel_torque(velocities, accelerations, road_loads, vehicle_mass, f, r_dynamic):
    """
    Calculates the wheel power [kW].

    :param velocities:
        Velocity [km/h].
    :type velocities: numpy.array | float

    :param accelerations:
        Acceleration [m/s2].
    :type accelerations: numpy.array | float

    :param road_loads:
        Cycle road loads [N, N/(km/h), N/(km/h)^2].
    :type road_loads: list, tuple

    :param vehicle_mass:
        Vehicle mass [kg].
    :type vehicle_mass: float

    :param f:
        Tyre rolling resistance.
    :type f: float

    :param r_dynamic
        Tyre radius.
    :type r_dynamic: float

    :return:
        Power at wheels [kW].
    :rtype: numpy.array | float
    """

    f0, f1, f2 = road_loads

    quadratic_term = f0 * math.cos(f) + (f1 + f2 * velocities) * velocities

    vel = velocities / 3600

    return (quadratic_term + 1.03 * vehicle_mass * accelerations + vehicle_mass * math.sin(f)) * r_dynamic


def calculate_wheel_speeds(velocities, r_dynamic):
    """
    Calculates rotating speed of the wheels [RPM].

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array | float

    :param r_dynamic:
        Dynamic radius of the wheels [m].
    :type r_dynamic: float

    :return:
        Rotating speed of the wheel [RPM].
    :rtype: numpy.array | float
    """

    return velocities * (30.0 / (3.6 * math.pi * r_dynamic))


def calculate_wheel_torques(wheel_powers, wheel_speeds):
    """
    Calculates torque at the wheels [N*m].

    :param wheel_powers:
        Power at the wheels [kW].
    :type wheel_powers: numpy.array | float

    :param wheel_speeds:
        Rotating speed of the wheel [RPM].
    :type wheel_speeds: numpy.array | float

    :return:
        Torque at the wheels [N*m].
    :rtype: numpy.array | float
    """

    pi = math.pi
    if isinstance(wheel_speeds, numpy.ndarray):
        return numpy.nan_to_num(wheel_powers / wheel_speeds * (30000.0 / pi))
    return wheel_powers / wheel_speeds * (30000.0 / pi) if wheel_speeds else 0.0


def calculate_final_drive_speeds_in(
        final_drive_speeds_out, final_drive_ratio_vector):
    """
    Calculates final drive speed [RPM].

    :param final_drive_speeds_out:
        Rotating speed of the wheel [RPM].
    :type final_drive_speeds_out: numpy.array | float

    :param final_drive_ratio_vector:
        Final drive ratio vector [-].
    :type final_drive_ratio_vector: numpy.array | float

    :return:
        Final drive speed in [RPM].
    :rtype: numpy.array | float
    """

    return final_drive_speeds_out * final_drive_ratio_vector


def calculate_final_drive_torque_losses_v1(
        n_wheel_drive, final_drive_torques_out, final_drive_ratio_vector,
        final_drive_efficiency):
    """
    Calculates final drive torque losses [N*m].

    :param n_wheel_drive:
        Number of wheel drive [-].
    :type n_wheel_drive: int

    :param final_drive_torques_out:
        Torque at the wheels [N*m].
    :type final_drive_torques_out: numpy.array | float

    :param final_drive_ratio_vector:
        Final drive ratio vector [-].
    :type final_drive_ratio_vector: numpy.array | float

    :param final_drive_efficiency:
        Final drive efficiency [-].
    :type final_drive_efficiency: float

    :return:
        Final drive torque losses [N*m].
    :rtype: numpy.array | float
    """

    eff_fd = final_drive_efficiency - (n_wheel_drive - 2) / 100
    to = final_drive_torques_out
    return (1 - eff_fd) / (eff_fd * final_drive_ratio_vector) * to


def calculate_final_drive_torques_in(
        final_drive_torques_out, final_drive_ratio_vector,
        final_drive_torque_losses):
    """
    Calculates final drive torque [N*m].

    :param final_drive_torques_out:
        Torque at the wheels [N*m].
    :type final_drive_torques_out: numpy.array | float

    :param final_drive_ratio_vector:
        Final drive ratio vector [-].
    :type final_drive_ratio_vector: numpy.array | float

    :param final_drive_torque_losses:
        Final drive torque losses [N*m].
    :type final_drive_torque_losses: numpy.array | float

    :return:
        Final drive torque in [N*m].
    :rtype: numpy.array | float
    """

    t = final_drive_torques_out / final_drive_ratio_vector

    return t + final_drive_torque_losses


def calculate_gear_box_speeds_in_v1(
        gear, final_drive_speed, gear_box_ratios, clutch):
    """
    Calculates Gear box speed vector [RPM].

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param gear_box_speeds_out:
        Wheel speed vector [RPM].
    :type gear_box_speeds_out: numpy.array

    :param gear_box_ratios:
        Gear box ratios [-].
    :type gear_box_ratios: dict

    :return:
        Gear box speed vector [RPM].
    :rtype: numpy.array
    """

    # d = {0: 0.0}
    #
    # d.update(gear_box_ratios)

    # ratios = numpy.vectorize(d.get)(gears)
    if clutch == 1:
        return 0

    return final_drive_speed * gear_box_ratios[gear - 1]


def gear_box_torques_in(
        min_engine_on_speed, gear_box_torques_out, gear_box_speeds_in,
        gear_box_speeds_out, par, clutch):
    """
    Calculates torque required according to the temperature profile [N*m].

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :param gear_box_torques_out:
        Torque gear_box vector [N*m].
    :type gear_box_torques_out: numpy.array

    :param gear_box_speeds_in:
        Engine speed vector [RPM].
    :type gear_box_speeds_in: numpy.array

    :param gear_box_speeds_out:
        Wheel speed vector [RPM].
    :type gear_box_speeds_out: numpy.array

    :param gear_box_efficiency_parameters_cold_hot:
        Parameters of gear box efficiency model:

            - `gbp00`,
            - `gbp10`,
            - `gbp01`
    :type gear_box_efficiency_parameters_cold_hot: dict

    :return:
        Torque required vector [N*m].
    :rtype: numpy.array
    """
    tgb, es, ws = gear_box_torques_out, gear_box_speeds_in, gear_box_speeds_out

    if gear_box_speeds_in != 0:
        ratio = gear_box_speeds_out / gear_box_speeds_in
    else:
        ratio = 0

    if tgb >= 0:
        if (es == 0 or ws == 0 or clutch == 1):
            res = 0
        else:
            res = (tgb * ratio - par['gbp10'] * es - par['gbp00']) / par['gbp01']
    else:
        res = -(-tgb * ratio * par['gbp01'] + par['gbp10'] * ws + par['gbp00'])

    return res


def create_gearbox_params(params, max_engine_torque):
    gbp10 = params['gbp10']['m'] * max_engine_torque + params['gbp10']['q']
    gbp00 = params['gbp00']['m'] * max_engine_torque + params['gbp00']['q']
    gbp01 = params['gbp01']['q']
    par = {'gbp00': gbp00, 'gbp10': gbp10, 'gbp01': gbp01}
    return par


def calculate_brake_mean_effective_pressures(
        engine_speeds_out, engine_powers_out, engine_capacity,
        min_engine_on_speed):
    """
    Calculates engine brake mean effective pressure [bar].

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param engine_powers_out:
        Engine power vector [kW].
    :type engine_powers_out: numpy.array

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :return:
        Engine brake mean effective pressure vector [bar].
    :rtype: numpy.array
    """

    if engine_speeds_out > min_engine_on_speed:
        # res = float(engine_powers_out * 2000) / (engine_capacity * 0.1 * engine_speeds_out / 60)
        res = (engine_powers_out / engine_speeds_out) * 1200000.0 / engine_capacity
    else:
        res = 0
    #
    # p = numpy.zeros_like(engine_powers_out)
    # p[b] = engine_powers_out[b] / engine_speeds_out[b]
    # p[b] *= 1200000.0 / engine_capacity

    return res


def mean_piston_speed(engine_speed, fuel_engine_stroke):
    # stroke is in mm
    # engine_speed is in RPM

    res = engine_speed * fuel_engine_stroke / 30000.0
    return res


def parameters(max_power, capacity, eng_fuel, fuel_turbo):
    if eng_fuel == 'diesel':
        params = {
            'a': -0.0005 * max_power + 0.438451,
            'b': -0.26503 * (-0.0005 * max_power + 0.43845) + 0.12675,
            'c': -0.08528 * (-0.26503 * (-0.0005 * max_power + 0.43845) + 0.12675) + 0.0003,
            'a2': -0.0012,
            'b2': 0,
            'l': -1.55291,
            'l2': -0.0076,
            't': 3.5,  # 2.7,
            'trg': 85.
        }
    elif fuel_turbo == 'yes':
        params = {
            'a': 0.8882 * max_power / capacity + 0.377,
            'b': -0.17988 * (0.882 * max_power / capacity + 0.377) + 0.0899,
            'c': -0.06492 * (-0.17988 * (0.882 * max_power / capacity + 0.377) + 0.0899) + 0.000117,
            'a2': -0.00385,
            'b2': 0,
            'l': -2.14063,
            'l2': -0.00286,
            't': 3.5,  # 2.7,
            'trg': 85.
        }
    else:
        params = {
            'a': 0.8882 * max_power / capacity + 0.377,
            'b': -0.17988 * (0.882 * max_power / capacity + 0.377) + 0.0899,
            'c': -0.06492 * (-0.17988 * (0.882 * max_power / capacity + 0.377) + 0.0899) + 0.000117,
            'a2': -0.00266,
            'b2': 0,
            'l': -2.49882,
            'l2': -0.0025,
            't': 3.5,  # 2.7,
            'trg': 85.
        }

    return params


# def calculate_p0(params, engine_capacity, engine_stroke,
#                  idle_engine_speed_median, engine_fuel_lower_heating_value):
#     """
#     Calculates the engine power threshold limit [kW].
#
#     :param params:
#         CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg).
#
#         The missing parameters are set equal to zero.
#     :type params: dict
#
#     :param engine_capacity:
#         Engine capacity [cm3].
#     :type engine_capacity: float
#
#     :param engine_stroke:
#         Engine stroke [mm].
#     :type engine_stroke: float
#
#     :param idle_engine_speed_median:
#         Engine speed idle median [RPM].
#     :type idle_engine_speed_median: float
#
#     :param engine_fuel_lower_heating_value:
#         Fuel lower heating value [kJ/kg].
#     :type engine_fuel_lower_heating_value: float
#
#     :return:
#         Engine power threshold limit [kW].
#     :rtype: float
#     """
#
#     engine_cm_idle = idle_engine_speed_median * engine_stroke / 30000.0
#
#     lhv = engine_fuel_lower_heating_value
#     FMEP = _calculate_fuel_mean_effective_pressure
#
#     engine_wfb_idle, engine_wfa_idle = FMEP(params, engine_cm_idle, 0, 1)
#     engine_wfa_idle = (3600000.0 / lhv) / engine_wfa_idle
#     engine_wfb_idle *= (3.0 * engine_capacity / lhv * idle_engine_speed_median)
#
#     return -engine_wfb_idle / engine_wfa_idle

def calculate_fuel_ABC(params, mean_piston_speed, n_powers, n_temperatures):
    p = params
    A = p['a2'] + p['b2'] * mean_piston_speed
    B = p['a'] + (p['b'] + p['c'] * mean_piston_speed) * mean_piston_speed
    C = numpy.power(n_temperatures, 0) * (p['l'] + p['l2'] * mean_piston_speed ** 2)
    C -= n_powers

    return A, B, C


def calculate_VMEP(A, B, C):
    if (B ** 2 - 4 * A * C) < 0:
        print('Negative (B ** 2 - 4 * A * C) ')
    if A != 0 and (B ** 2 - 4 * A * C) >= 0:
        res = (-B + math.sqrt(B ** 2 - 4 * A * C)) / (2 * A)
    else:
        res = float(-C) / B
    return res


def calc_fuel_consumption(VMEP, engine_capacity, engine_fuel_lower_heating_value, engine_rpm, sim_step):
    # fuel consumption in grams per time step
    res = sim_step * ((VMEP * engine_capacity * 0.1 * (float(engine_rpm) / 60)) / (2 * engine_fuel_lower_heating_value))
    if res < 0:
        return 0
    return res



from os import path as osp, chdir
import matplotlib.pyplot as plt
from co2mpas_driver import dsp as driver

my_dir = osp.dirname(osp.abspath(__file__))
chdir(my_dir)


def simple_run():
    """
    This example plots the acceleration and deceleration potential curves (over speed) of the vehicle for each gear.

    :return:
    """
    # A sample car id from the database = {
    car_id_db = {
        'fuel_engine': [35135, 39393, 27748, 8188, 40516, 35452, 40225, 7897, 7972, 41388, 5766, 9645, 9639, 5798, 8280,
                        34271, 34265, 6378, 39723, 34092, 2592, 5635, 5630, 7661, 7683, 8709, 9769, 1872, 10328, 35476,
                        41989, 26799, 26851, 27189, 23801, 3079, 36525, 47766, 6386, 33958, 33986, 5725, 5718, 36591,
                        4350, 39396, 40595, 5909, 5897, 5928, 5915, 40130, 42363, 34760, 34766, 1835, 36101, 42886,
                        1431, 46547, 44799, 41045, 39820, 34183, 34186, 20612, 20605, 1324, 9882, 4957, 5595, 18831,
                        18833, 9575, 5380, 9936, 7995, 6331, 18173, 34286, 34279, 20706, 34058, 34057, 24268, 19028,
                        19058, 7979, 22591, 34202, 40170, 44599, 5358, 5338, 34015, 9872, 9856, 6446, 8866, 9001, 9551,
                        6222],
        'electric_engine': [47844]
    }
    car_id = 35135

    # Core model, this will select and execute the proper functions for the given inputs and returns the output
    # You can also pass your vehicle's database path db_path='path to vehicle db'
    if car_id in car_id_db['electric_engine']:
        sol = driver(dict(vehicle_id=47844))['outputs']
    else:
        # The gear shifting style as described in the TRR paper.
        gs_style = 0.8
        # 4th degree polynomial
        degree = 4
        sol = driver(dict(vehicle_id=car_id, inputs=dict(inputs=dict(
            gear_shifting_style=gs_style, degree=degree, use_linear_gs=True,
            use_cubic=False))))['outputs']
    # Plots workflow of the core model, this will automatically open an internet browser and shows the work flow
    # of the core model. you can click all the rectangular boxes to see in detail sub models like load, model,
    # write and plot.
    # you can uncomment to plot the work flow of the model
    # driver.plot(1)  # Note: run your IDE as Administrator if file permission error.

    fig = plt.figure()
    for gear, curve in enumerate(sol['discrete_acceleration_curves'], start=1):
        sp_bins = list(curve['x'])
        acceleration = list(curve['y'])
        plt.plot(sp_bins, acceleration, label=f'gear {gear}')
    plt.legend()
    plt.text(27, 2.5, f'Simulation car:{car_id}')
    fig.suptitle('Acceleration over speed', x=0.54, y=1, fontsize=12)
    plt.xlabel('Speed (m/s)', fontsize=12)
    plt.ylabel('Acceleration (m/s2)', fontsize=12)
    plt.grid()
    #fig.savefig(f'sample_4degree_with_linear_Acceleration_Speed_for_vehicle_id_{car_id}_gear_shifting_style_{gs_style}.jpg')

    fig1 = plt.figure('Deceleration over Speed')
    for gear, curve in enumerate(sol['discrete_deceleration_curves'], start=1):
        sp_bins = list(curve['x'])
        deceleration = list(curve['y'])
        plt.plot(sp_bins, deceleration, label=f'gear {gear}')
    plt.legend()
    plt.text(23, 2.5, f'Simulation car:{car_id}')
    fig1.suptitle('Deceleration over Speed', x=0.54, y=1, fontsize=12)
    plt.xlabel('Speed (m/s)', fontsize=12)
    plt.ylabel('Deceleration (m/s2)', fontsize=12)
    plt.grid()
    #fig1.savefig(f'sample_4degree_with_linear_Deceleration_Speed_for_vehicle_id_{car_id}_gear_shifting_style_{gs_style}.jpg')

    plt.show()

    return 0


if __name__ == '__main__':
    simple_run()



# -*- coding: utf-8 -*-
#
# Copyright 2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl


class Driver:
    """
        Blueprint for driver.
    """

    def __init__(self, vehicle_mass, car_type, final_drive, gearbox_type, max_power,
                 fuel_type, type_of_car, car_width, car_height, transmission, gs, curves,
                 curves_dec, driver_style, r_dynamic=None, gear_box_ratios=None,
                 engine_max_torque=None, fuel_eng_capacity=None, fuel_engine_stroke=None,
                 fuel_turbo=None):
        self.vehicle_mass = vehicle_mass
        self.r_dynamic = r_dynamic
        self.car_type = car_type
        self.final_drive = final_drive
        self.gear_box_ratios = gear_box_ratios
        self.gearbox_type = gearbox_type
        self.engine_max_torque = engine_max_torque
        self.fuel_eng_capacity = fuel_eng_capacity
        self.max_power = max_power
        self.fuel_engine_stroke = fuel_engine_stroke
        self.fuel_type = fuel_type
        self.type_of_car = type_of_car
        self.car_width = car_width
        self.car_height = car_height
        self.fuel_turbo = fuel_turbo

        self.transmission = transmission
        self.gs = gs
        self.curves = list(curves)
        self.curves_dec = list(curves_dec)
        self.driver_style = driver_style
        self._velocity = self._gear_count = self._gear = None

    def reset(self, starting_velocity):
        from .simulation import gear_for_speed_profiles as func
        self._gear, self._gear_count = func(self.gs, starting_velocity, 0, 0)
        self._gear_count, self._velocity = 0, starting_velocity
        return self

    def update(self, next_velocity, gear, gear_cnt, sim_step):
        from .simulation import gear_for_speed_profiles as func
        g, gc = func(self.gs, next_velocity, gear, gear_cnt)
        cnt = int(divmod(10 * 0.5, 10 * sim_step)[0])
        is_between = 0 < gear_cnt < cnt
        if g == gear and is_between:
            gc += 1
        elif g == gear and gear_cnt in [0, cnt]:
            gc = gear_cnt
        self._gear, self._gear_count, self._velocity = g, gc, next_velocity
        return g, gc

    def __call__(self, dt, desired_velocity, update=True):
        from .simulation import (
            gear_for_speed_profiles, accMFC, correct_acc_clutch_on
        )
        g, gc = gear_for_speed_profiles(
            self.gs, self._velocity, self._gear, self._gear_count
        )

        a = correct_acc_clutch_on(gc, accMFC(
            self._velocity, self.driver_style, desired_velocity,
            self.curves[g - 1], self.curves_dec[g - 1]
        ), self.transmission)
        v = self._velocity + a * dt
        if update:
            self._gear, self._gear_count, self._velocity = g, gc, v
        return g, gc, v, a

    def calculate_fuel_consumption(self, speed, acceleration, gear, gear_count, sim_step, slope=0):
        import math
        from co2mpas_driver.common import functions as func, vehicle_specs_class as vcc
        from .co2mpas import estimate_f_coefficients

        if self.fuel_type == 'electricity':
            return 0

        road_loads = estimate_f_coefficients(self.vehicle_mass, self.type_of_car, self.car_width,
                                             self.car_height)

        hardcoded_params = vcc.hardcoded_params()
        if self.gearbox_type == 'manual':
            veh_params = hardcoded_params.params_gearbox_losses['Manual']
            gb_type = 0
        else:
            veh_params = hardcoded_params.params_gearbox_losses['Automatic']
            gb_type = 1

        n_wheel_drive = self.car_type

        # The power on wheels in kW
        veh_wheel_power = func.calculate_wheel_power(speed, acceleration, road_loads, self.vehicle_mass, slope)

        # The speed on the wheels in [RPM]
        veh_wheel_speed = func.calculate_wheel_speeds(speed, self.r_dynamic)

        # # The torque on the wheels in [N*m]
        veh_wheel_torque = func.calculate_wheel_torques(veh_wheel_power, veh_wheel_speed)

        # Calculates final drive speed in RPM
        final_drive_speed = func.calculate_final_drive_speeds_in(veh_wheel_speed, self.final_drive)

        # Final drive torque losses [N*m].
        final_drive_torque_losses = func.calculate_final_drive_torque_losses_v1(n_wheel_drive, veh_wheel_torque,
                                                                                self.final_drive,
                                                                                hardcoded_params.final_drive_efficiency)

        # Final drive torque in [N*m].
        final_drive_torque_in = func.calculate_final_drive_torques_in(veh_wheel_torque, self.final_drive,
                                                                      final_drive_torque_losses)

        gear_box_speeds_in = func.calculate_gear_box_speeds_in_v1(gear, final_drive_speed, self.gear_box_ratios, 0)

        gearbox_params = func.create_gearbox_params(veh_params, self.engine_max_torque)

        gear_box_torques_in = func.gear_box_torques_in(hardcoded_params.min_engine_on_speed, final_drive_torque_in,
                                                       gear_box_speeds_in, final_drive_speed, gearbox_params,
                                                       gear_count)

        gear_box_power_out = 2 * math.pi * gear_box_torques_in * gear_box_speeds_in / 60000

        # gear_box_power_out = func.calculate_gear_box_power_out(gear_box_torques_in,
        #                                                        gear_box_speeds_in)

        br_eff_pres = func.calculate_brake_mean_effective_pressures(gear_box_speeds_in, gear_box_power_out,
                                                                    self.fuel_eng_capacity,
                                                                    hardcoded_params.min_engine_on_speed)

        engine_cm = func.mean_piston_speed(gear_box_speeds_in, self.fuel_engine_stroke)

        params = func.parameters(self.max_power, self.fuel_eng_capacity, self.fuel_type, self.fuel_turbo)
        fuel_A, fuel_B, fuel_C = func.calculate_fuel_ABC(params, engine_cm, br_eff_pres, 100)

        if br_eff_pres > 20:
            # Control for unrealistic Break Mean Effective Pressure values.
            print('BMEP> %.2f bar, EngineCM: %.2f, Gear: %d : Check out the MFC output. The engine will blow up!!!!' % (
                br_eff_pres, engine_cm, gear))

        if br_eff_pres > -0.5:
            # Fuel mean effective pressure
            VMEP = func.calculate_VMEP(fuel_A, fuel_B, fuel_C)
        else:
            VMEP = 0
        lower_heating_value = hardcoded_params.LHV[self.fuel_type]

        # Fuel consumption in grams.
        fc = func.calc_fuel_consumption(VMEP, self.fuel_eng_capacity,
                                        lower_heating_value, gear_box_speeds_in,
                                        sim_step)

        return fc



# Computation of the MFC vehicle acceleration - speed curve.
# coding=utf-8
import functools as functools
import numpy as np
from scipy.interpolate import CubicSpline, interp1d
from co2mpas_driver.common import defaults as defaults


def get_full_load(ignition_type):
    """
    Returns vehicle full load curve.

    :param ignition_type:
        Engine ignition type (positive or compression).
    :type ignition_type: str

    :return:
        Vehicle normalized full load curve.
    :rtype: scipy.interpolate.InterpolatedUnivariateSpline
    """

    xp, fp = defaults.dfl.functions.get_full_load.FULL_LOAD[ignition_type]
    func = functools.partial(
        np.interp, xp=xp, fp=fp, left=fp[0], right=fp[-1]
    )

    return func


def calculate_full_load_speeds_and_powers(full_load_curve, my_car):
    """
    Calculates the full load speeds and powers [RPM, kW].

    :param full_load_curve:
        Vehicle normalized full load curve.
    :type full_load_curve: scipy.interpolate.InterpolatedUnivariateSpline

    :param engine_max_power:
        Engine nominal power [kW].
    :type engine_max_power: float

    :param engine_max_speed_at_max_power:
        Engine nominal speed at engine nominal power [RPM].
    :type engine_max_speed_at_max_power: float

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :return:
         T1 map speed [RPM] and power [kW] vectors.
    :rtype: (numpy.array, numpy.array)
    """

    n_norm = np.arange(0.0, 1.21, 0.1)
    full_load_powers = full_load_curve(n_norm) * my_car.engine_max_power
    idle = my_car.idle_engine_speed[0]
    full_load_speeds = n_norm * (my_car.engine_max_speed_at_max_power - idle) + idle

    return full_load_speeds, full_load_powers


# The maximum force that the vehicle can have on the road
def Armax(my_car, road_type=1):
    """
    Calculating the maximum acceleration possible for the vehicle object my_car,
    under road_type conditions.

    :param my_car: vehicle specs object
    :param road_type: road condition (1: normal, 2: wet, 3: icy)
    :return:
    """
    if my_car.car_type == 2:  # forward-wheel drive vehicles
        fmass = 0.6 * my_car.veh_mass
    elif my_car.car_type == 4:  # rear-wheel drive vehicles
        fmass = 0.45 * my_car.veh_mass
    else:  # all-wheel drive vehicles, 4x4
        fmass = 1 * my_car.veh_mass

    if road_type == 1:
        mh_base = 0.75  # for normal road
    elif road_type == 2:
        mh_base = 0.25  # for wet road
    else:
        mh_base = 0.1  # for icy road
    # Optimal values:
    # 0.8 dry, 0.6 light rain, 0.4 heavy rain, 0.1 icy
    alpha = 43.398
    beta = 5.1549
    mh = mh_base * (alpha * np.log(my_car.engine_max_power) + beta) / 190

    # * cos(f) for the gradient of the road. Here we consider as 0
    Frmax = fmass * 9.8066 * mh

    return Frmax / my_car.veh_mass


# Calculates a spline with the resistances
def veh_resistances(f0, f1, f2, sp, total_mass):
    """
    Return the resistances that a vehicle faces, per speed

    :param f0:
    :type f0: float
    :param f1:
    :type f1: float
    :param f2:
    :type f2:float
    :param sp:
    :type sp: list
    :param total_mass:
    :type total_mass: float
    :return: resistance_spline_curve, resistance_spline_curve_f
    :rtype: CubicSpline, CubicSpline
    """
    Fresistance = []
    for i in range(len(sp)):
        Fresistance.append(f0 + f1 * sp[i] * 3.6 + f2 * pow(sp[i] * 3.6, 2))
        # Facc = Fmax @ wheel - f0 * cos(a) - f1 * v - f2 * v2 - m * g * sin(a)

    aprx_mass = int(total_mass)
    Aresistance = [x / aprx_mass for x in Fresistance]
    a = int(np.floor(sp[0]))
    b = int(np.floor(sp[-1]))
    resistance_spline_curve = CubicSpline([k for k in range(a - 10, a)] + sp + [k for k in range(b + 1, b + 11)], \
                                          [Aresistance[0]] * 10 + Aresistance + [Aresistance[-1]] * 10)
    resistance_spline_curve_f = CubicSpline([k for k in range(a - 10, a)] + sp + [k for k in range(b + 1, b + 11)],
                                            [Fresistance[0]] * 10 + Fresistance + [Fresistance[-1]] * 10)

    return resistance_spline_curve, resistance_spline_curve_f


def estimate_f_coefficients(my_car, passengers=0):
    """
    f0, f1, f2 coefficients of resistance are estimated

    :param my_car:
    :param passengers:
    :type passengers: int
    :return:
    """

    d = {}
    # Fill in the entries one by one
    d["cabriolet"] = 0.28
    d["sedan"] = 0.27
    d["hatchback"] = 0.3
    d["stationwagon"] = 0.28
    d["suv/crossover"] = 0.35
    d["mpv"] = 0.3
    d["coupé"] = 0.27
    d["pick-up"] = 0.4

    rolling_res_coef = 0.009  # Constant for the moment
    theor_aero_coeff = d[my_car.type_of_car]

    operating_mass = my_car.veh_mass + 100 + 75 * passengers
    f0 = (operating_mass + 100) * rolling_res_coef * 9.81
    f2 = 0.5 * 1.2 * (0.84 * my_car.car_width * my_car.car_height * theor_aero_coeff) / pow(3.6, 2)
    f1 = -71.735 * f2 + 2.7609

    return f0, f1, f2


def get_load_speed_n_torque(my_car):
    """
    :param my_car:
    :return:
        full_load_speeds, full_load_torques
    :rtype: numpy.array, numpy.array
    """

    full_load = get_full_load(my_car.ignition_type)
    full_load_speeds, full_load_powers = calculate_full_load_speeds_and_powers(full_load, my_car)
    full_load_torques = full_load_powers * 1000 * (full_load_speeds / 60 * 2 * np.pi) ** -1
    return full_load_speeds, full_load_torques


def get_speeds_n_accelerations_per_gear(my_car, full_load_speeds,
                                        full_load_torques):
    """
    Speed and acceleration points per gear are calculated based on
    full load curve, new version works with array and
    forbid acceleration over the maximum vehicle speed

    :param my_car:
    :param full_load_speeds:
    :type full_load_speeds: numpy.array
    :param full_load_torques:
    :type full_load_torques: numpy.array
    :return:
        speed_per_gear, acc_per_gear
    :rtype: list, list
    """
    speed_per_gear, acc_per_gear = [], []

    full_load_speeds = np.array(full_load_speeds)
    full_load_torques = np.array(full_load_torques)

    for j in range(len(my_car.gr)):
        mask = full_load_speeds > 1.25 * my_car.idle_engine_speed[0]

        temp_speed = 2 * np.pi * my_car.tire_radius * full_load_speeds[mask] * (1 - my_car.driveline_slippage) / (
            60 * my_car.final_drive * my_car.gr[j])
        speed_per_gear.append(temp_speed)

        temp_acc = full_load_torques[mask] * (my_car.final_drive * my_car.gr[j]) * my_car.driveline_efficiency / (
            my_car.tire_radius * my_car.veh_mass)

        acc_per_gear.append(temp_acc)

    return speed_per_gear, acc_per_gear


def get_cubic_splines_of_speed_acceleration_relationship(my_car, speed_per_gear, acc_per_gear):
    """
    Based on speed/acceleration points per gear, cubic splines are calculated
    (old MFC)

    :param my_car:
    :param speed_per_gear:
    :param acc_per_gear:
    :return:
    """
    cs_acc_per_gear = []
    for j in range(len(my_car.gr)):
        # cs_acc_per_gear.append([])
        a = np.round((speed_per_gear[j][0]), 2) - 0.01
        b = np.round((speed_per_gear[j][-1]), 2) + 0.01
        prefix_list = [a - k * 0.1 for k in range(10, -1, -1)]
        suffix_list = [b + k * 0.1 for k in range(0, 11, 1)]
        cs_acc_per_gear.append(CubicSpline(
            prefix_list + list(speed_per_gear[j]) + suffix_list,
            [acc_per_gear[j][0]] * len(prefix_list) + list(acc_per_gear[j]) + [
                acc_per_gear[j][-1]] * len(suffix_list))
        )

    return cs_acc_per_gear


def get_start_stop(my_car, speed_per_gear, acc_per_gear, poly_spline):
    """

    Calculate Speed boundaries for each gear

    :param my_car:
    :param speed_per_gear:
    :type speed_per_gear: list
    :param acc_per_gear:
    :type acc_per_gear: list
    :param poly_spline:
    :type poly_spline: list
    :return: Start, Stop
    :rtype: list, list
    """
    import copy
    speed_per_gear = copy.deepcopy(speed_per_gear)
    acc_per_gear = copy.deepcopy(acc_per_gear)
    # To ensure that a higher gear starts from higher speed
    for j in range(len(my_car.gr) - 1, 0, -1):
        for k in range(len(speed_per_gear[j])):
            if speed_per_gear[j - 1][0] < speed_per_gear[j][0]:
                break
            else:
                # If the gear ratios are not declining,
                # there is an error in the database. Return error.
                return
                # speed_per_gear[j] = speed_per_gear[j][3:]

    # Find where the curve of each gear cuts the next one.
    for j in range(len(my_car.gr) - 1):
        for k in range(np.minimum(len(speed_per_gear[j]), len(speed_per_gear[j + 1]))):
            if (speed_per_gear[j][k] > speed_per_gear[j + 1][0]) & (
                        poly_spline[j + 1](speed_per_gear[j][k]) > poly_spline[j](speed_per_gear[j][k])):
                max_point = k
                speed_per_gear[j] = speed_per_gear[j][:max_point]
                acc_per_gear[j] = acc_per_gear[j][:max_point]
                break

    # The limits of the gears that should be provided to the gear shifting model
    Start = []
    Stop = []
    for i in speed_per_gear:
        Start.append(i[0])
        Stop.append(min(i[-1], my_car.veh_max_speed))
    Start[0] = 0
    return Start, Stop


def get_resistances(my_car, sp_bins):
    """
    Calculate resistances and return spline.

    :param my_car:
    :param sp_bins:
    :type sp_bins: np.array
    :return: car_res_curve, car_res_curve_force, Alimit
    :rtype: CubicSpline, CubicSpline, float
    """
    f0, f1, f2 = estimate_f_coefficients(my_car, 0)
    car_res_curve, car_res_curve_force = veh_resistances(f0, f1, f2, list(sp_bins), my_car.veh_mass)
    Alimit = Armax(my_car)
    return car_res_curve, car_res_curve_force, Alimit


def calculate_curves_to_use(poly_spline, Start, Stop, Alimit, car_res_curve,
                            sp_bins):
    """

    Get the final speed acceleration curves based on full load curves and resistances for all curves

    :param poly_spline:
    :type poly_spline: list
    :param Start:
    :type Start: list
    :param Stop:
    :type Stop: limit
    :param Alimit:
    :type Alimit: float
    :param car_res_curve:
    :type car_res_curve: CubicSpline
    :param sp_bins:
    :type sp_bins: numpy.array
    :return: Curves
    :rtype: list
    """
    Res = []

    for gear, acc in enumerate(poly_spline):
        start = Start[gear] * 0.9
        stop = Stop[gear] + 0.1

        final_acc = acc(sp_bins) - car_res_curve(sp_bins)
        final_acc[final_acc > Alimit] = Alimit

        final_acc[(sp_bins < start)] = 0
        final_acc[(sp_bins > stop)] = 0
        final_acc[final_acc < 0] = 0

        Res.append(interp1d(sp_bins, final_acc))

    return Res


def calculate_dec_curves_to_use(sp_bins):
    '''
    Calculates deceleration curves.
    :param sp_bins:
    :return:
    '''
    ppar = [0.0045, -0.1710, -1.8835]
    dec_curves = np.poly1d(ppar)
    final_dec = []
    Curves_dec = []
    for k in range(len(sp_bins)):
        final_dec.append(dec_curves(sp_bins[k]))
    from scipy.interpolate import interp1d
    Curves_dec.append(interp1d(sp_bins, final_dec))

    return Curves_dec


def get_starting_speed(speed_per_gear):
    starting_speed = speed_per_gear[0][0]
    return starting_speed


def get_tan_coefs(speed_per_gear, acc_per_gear, degree):
    """
    Full load curve is fitted to a polynomial of degree

    :param speed_per_gear:
    :type speed_per_gear: list
    :param acc_per_gear:
    :type acc_per_gear: list
    :param degree:
    :return: coefs_per_gear:
        the coefficients of the polynomial for each gear
    :rtype: list
    """

    coefs_per_gear = []
    for speeds, acceleration in zip(speed_per_gear, acc_per_gear):
        fit_coef = np.polyfit(speeds, acceleration, degree)
        coefs_per_gear.append(fit_coef)

    return coefs_per_gear


def get_spline_out_of_coefs(coefs_per_gear, starting_speed):
    """
    Use the coefs to get a "spline" that could be used.
    AT TIME IT IS USED AS calculate_curve_to_use FUNCTION IS USING SPLINES

    :param coefs_per_gear:
    :type coefs_per_gear: list
    :param starting_speed:
    :type starting_speed: float
    :return:
        spline_from_poly
    :rtype: list
    """

    degree = len(coefs_per_gear[0]) - 1
    vars = np.arange(degree, -1, -1)

    spline_from_poly = []

    """
    For the first gear, some points are added at the beginning to avoid
    unrealistic drops 
    """
    x_new = np.insert(np.arange(starting_speed, 70, 0.1), [0, 0],
                      [0, starting_speed / 2])
    a_new = np.array(
        [np.dot(coefs_per_gear[0], np.power(i, vars)) for i in x_new])
    a_new[0] = a_new[2]
    a_new[1] = a_new[2]
    spline_from_poly.append(interp1d(x_new, a_new, fill_value='extrapolate'))

    for fit_coef in coefs_per_gear[1:]:
        x_new = np.arange(0, 70, 0.1)
        a_new = np.array([np.dot(fit_coef, np.power(i, vars)) for i in x_new])
        spline_from_poly.append(interp1d(x_new, a_new, fill_value='extrapolate'))

    return spline_from_poly


def ev_curve(my_car):
    """
    Full load curve of EVs (based on Yinglong)

    :param my_car:
    :return:
    """

    veh_max_acc = my_car.motor_max_torque * (
            my_car.final_drive * my_car.gr) * my_car.driveline_efficiency / (
                my_car.tire_radius * my_car.veh_mass)  # m/s2

    speeds = np.arange(0, my_car.veh_max_speed + 0.1, 0.1)  # m/s

    with np.errstate(divide='ignore'):
        accelerations = my_car.engine_max_power * 1000 * my_car.driveline_efficiency / (speeds * my_car.veh_mass)

    accelerations[accelerations > veh_max_acc] = veh_max_acc
    accelerations[accelerations < 0] = 0

    cs_acc_ev = CubicSpline(speeds, accelerations)
    start = 0
    stop = my_car.veh_max_speed

    return [cs_acc_ev], [start], [stop]



import matplotlib
from co2mpas_driver import dsp as driver
import matplotlib.pyplot as plt
import numpy as np
import math as m

matplotlib.use('qt4agg')


def idm_calc_vel(af_mfc, fol_v, lead_v, vn, dist, bn, an, sim_step, **kwargs):
    if 'ds' in kwargs:  # inter vehicle spacing at a stop
        ds = kwargs['ds']
    else:
        ds = 2
    if 'th' in kwargs:  # the desired time headway to the vehicle in front
        th = kwargs['th']
    else:
        th = 1.5
    if 'delta' in kwargs:  # the desired time headway to the vehicle in front
        delta = kwargs['delta']
    else:
        delta = 4
    bn_idm = abs(bn)  # braking deceleration  b - a positive number
    return max(0, fol_v + (af_mfc + -an * (pow(max(0, (ds + fol_v * th) / dist + (fol_v * (fol_v - lead_v)) /
                                                   (2 * m.sqrt(an * bn_idm) * dist)), 2))) * sim_step)


# IDM params
amax = 3
bn = -2
delta = 4
ds = 2
th = 1.5

np.random.seed(123)

# veh_ids = [27748, 8188]
# veh_ids = [39393, 8188, 40516, 35452, 40225, 7897, 7972, 41388, 5766,
#            9645, 9639, 5798, 8280, 34271, 34265, 6378, 39723, 34092, 2592,
#            5635, 5630, 7661, 7683]
veh_pool = [39393, 8188, 40516, 35452, 40225, 7897, 7972, 41388, 5766,
            9645, 9639, 5798, 8280, 34271, 34265, 6378, 39723, 34092, 2592,
            5635, 5630, 7661, 7683]
num_vehs = 1
veh_ids = [np.random.choice(veh_pool) for i in range(num_vehs)]

v_des = 100 / 3.6
v_start = 20
dt = 0.1

sim_duration = 30  # sec
times = np.arange(0, sim_duration, dt)

vehicles = [driver(dict(vehicle_id=i, inputs=dict(inputs=dict(
    gear_shifting_style=1, driver_style=1,
    use_linear_gs=True,
    use_cubic=False))))['outputs'] for i in
            veh_ids]
models = [veh['driver_simulation_model'] for veh in vehicles]
dcurves = [veh['driver_simulation_model'] for veh in vehicles]

dist = 30  # meters

test_leader_sp_profile = [20 / 3.6] * int(10 * sim_duration / 6) + [50 / 3.6] * int(10 * sim_duration / 6) + [
    120 / 3.6] * int(10 * sim_duration / 6) + [80 / 3.6] * int(10 * sim_duration / 6) + [
                             60 / 3.6] * int(10 * sim_duration / 6) + [100 / 3.6] * int(10 * sim_duration / 6)
leader_dt = np.cumsum([i * dt for i in test_leader_sp_profile]) + dist * (len(models) + 1)

res = {}
for idx, myt in enumerate(times):
    if (len(times) - idx) % 100 == 0:
        print(len(times) - idx)
    for v_idx, my_veh in enumerate(models):
        if myt == times[0]:
            my_veh.reset(v_start)
            res[my_veh] = {'acc': [0], 'speed': [v_start], 'pos': [dist * (len(models) - v_idx)]}
            sl_prev = 20 / 3.6
            continue

        # desired speed in order not to cover the distance to the leader
        # distL is the position of the leader + velocity of leader multiplied by dt
        if v_idx == 0:
            distL = leader_dt[idx] - res[models[v_idx]]['pos'][idx - 1]
        else:
            distL = res[models[v_idx - 1]]['pos'][idx] - res[models[v_idx]]['pos'][idx - 1]
        v_desired_new = min(distL / dt, v_des)

        # Compute MFC gear, next_velocity, acceleration - The positing is not needed
        gear, gear_count, next_velocity, acc = my_veh(dt, v_desired_new)

        # Fuel consumption
        # fc = my_veh.calculate_fuel_consumption(next_velocity, acc, gear, gear_count, dt)

        final_acc = acc

        # So the next velocity will be based on the cf acceleration (not the MFC one)
        next_velocity = res[my_veh]['speed'][-1] + final_acc * dt

        ####
        if next_velocity < 0:
            next_velocity = 0
        if v_desired_new - 0.001 < next_velocity < v_desired_new + 0.001:
            next_velocity = v_desired_new
        if next_velocity > v_desired_new:
            next_velocity = v_desired_new
        ####

        # I save my results. Again here positing is of no use
        res[my_veh]['acc'].append(final_acc)
        res[my_veh]['speed'].append(next_velocity)
        if res[models[v_idx]]['pos'][idx - 1] + next_velocity * (dt) + 0.5 * acc * dt ** 2 <0:
            print('check')
        dtf = res[models[v_idx]]['pos'][idx - 1] + next_velocity * (dt) + 0.5 * acc * dt ** 2
        res[models[v_idx]]['pos'].append(dtf)

        # update the vehicle for the next MFC
        gear, gear_count = my_veh.update(next_velocity, gear, gear_count, dt)
        print(f':{gear_count}')

        # if (gear_count !=0 and v_idx == 0):
        #     fig = plt.figure()
        #     fig.suptitle(gear)
        #     for cgear, curve in enumerate(vehicles[v_idx]['discrete_deceleration_curves']):
        #         sp_bins = list(curve['x'])
        #         acceleration = list(curve['y'])
        #         plt.plot(sp_bins, acceleration, label=f'gear {gear}')
        #     plt.legend()
        #     plt.text(30, 2.5, f'Simulation car:{veh_ids}, GS:{1}')
        #     fig.suptitle('Acceleration over speed', x=0.54, y=1, fontsize=12)
        #     plt.xlabel('Speed (m/s)', fontsize=12)
        #     plt.ylabel('Acceleration (m/s2)', fontsize=12)
        #     plt.show()

fig = plt.figure('Trajectory')
ax = fig.add_subplot(111)
ax.plot(times, leader_dt, label='leader')
for idx, my_veh in enumerate(models):
    # plt.figure('Acceleration-Speed')
    # plt.plot(res[my_veh]['speed'], res[my_veh]['acc'], label=idx)
    # plt.plot(res[my_veh]['speed'], res[my_veh]['acc'], label=idx)
    # plt.legend()
    # plt.show()

    # plt.figure('Speed Profile')
    # plt.plot(times, test_leader_sp_profile, label='leader')
    # plt.plot(times, res[my_veh]['speed'], label=idx)
    # plt.legend()
    # plt.show()

    follower_dt = res[my_veh]['pos']

    ax.plot(times, follower_dt, label=idx)
    # plt.legend()
miny = max(res[models[-1]]['pos'])
maxy = max(res[models[0]]['pos'])
if miny < maxy:
    ax.set_ylim(miny, maxy)
plt.show()
